@@ -32,6 +32,7 @@ done
 
 repo_root="$(dirname "$0")/.."
 fish_site=$repo_root/../fish-site
+fish_site_repo=git@github.com:$repository_owner/fish-site
 
 for path in . "$fish_site"
 do
@@ -42,6 +43,13 @@ do
     fi
 done
 
+(
+    cd "$fish_site"
+    [ "$(git rev-parse HEAD)" = \
+        "$(git ls-remote "$fish_site_repo" refs/heads/master |
+            awk '{print $1}')" ]
+)
+
 if git tag | grep -qxF "$version"; then
     echo >&2 "$0: tag $version already exists"
     exit 1
@@ -147,15 +155,21 @@ rm -rf "$tmpdir"
     " | sed 's,^\s*| \?,,')"
 )
 
-# Approve macos-codesign
-# TODO what if current user can't approve?
-gh_pending_deployments() {
+gh_api_repo() {
+    path=$1
+    shift
     command gh api \
         -H "Accept: application/vnd.github+json" \
         -H "X-GitHub-Api-Version: 2022-11-28" \
-        "/repos/$repository_owner/fish-shell/actions/runs/$run_id/pending_deployments" \
+        "/repos/$repository_owner/fish-shell/$path" \
         "$@"
 }
+
+# Approve macos-codesign
+# TODO what if current user can't approve?
+gh_pending_deployments() {
+    gh_api_repo "actions/runs/$run_id/pending_deployments" "$@"
+}
 while {
     environment_id=$(gh_pending_deployments | jq .[].environment.id)
     [ -z "$environment_id" ]
@@ -170,7 +184,7 @@ echo '
             "comment": "Approved via ./build_tools/release.sh"
         }
     ' |
-gh_pending_deployments -XPOST --input=-
+gh_pending_deployments --method POST --input=-
 
 # Await completion.
 gh run watch "$run_id"
@@ -198,12 +212,12 @@ done
     " | sed 's,^\s*| \?,,')"
     # This takes care to support remote names that are different from
     # fish-shell remote name. Also, support detached HEAD state.
-    git push git@github.com:$repository_owner/fish-site HEAD:master
+    git push "$fish_site_repo" HEAD:master
 )
 
-if [ -n "$integration_branch" ]; then
+if [ -n "$integration_branch" ]; then {
     git push $remote "$version^{commit}":refs/heads/$integration_branch
-else
+} else {
     changelog=$(cat - CHANGELOG.rst <<EOF
 fish ?.?.? (released ???)
 =========================
@@ -213,23 +227,27 @@ EOF
     printf %s\\n "$changelog" >CHANGELOG.rst
     CommitVersion ${version}-snapshot "start new cycle"
     git push $remote HEAD:master
-fi
-
-# TODO This can currently require a TTY for editing and password
-# prompts.
-if [ "$repository_owner" = fish-shell ]; then {
-    mail=$(mktemp)
-    cat >$mail <<EOF
-From: $(git var GIT_AUTHOR_IDENT | sed 's/ [0-9]* +[0-9]*$//')
-Subject: fish $version released
-
-See https://github.com/fish-shell/fish-shell/releases/tag/$version
-EOF
-    git send-email --suppress-cc=all --confirm=always $mail \
-        --to="fish-users Mailing List <fish-users@lists.sourceforge.net>"
-    rm $mail
 } fi
 
+milestone_number=$(
+    gh_api_repo milestones?state=open |
+        jq '.[] | select(.title == "fish '"$version"'") | .number'
+)
+gh_api_repo --method PATCH milestones/$milestone_number \
+    --raw-field state=closed
+
+next_patch_version=$(
+    echo "$version" | awk -F. '
+        NF == 3 && $3 ~ /[0-9]+/ {
+            printf "%s.%s.%s", $1, $2, $3+1
+        }
+    '
+)
+if [ -n "$next_patch_version" ]; then
+    gh_api_repo --method POST milestones \
+        --raw-field title="fish $next_patch_version"
+fi
+
 exit
 
 }
