@@ -1,5 +1,7 @@
 //! The classes responsible for autoloading functions and completions.
 
+#[cfg(feature = "embed-data")]
+use crate::common::wcs2string;
 use crate::common::{escape, ScopeGuard};
 use crate::env::Environment;
 use crate::io::IoChain;
@@ -7,8 +9,12 @@ use crate::parser::Parser;
 #[cfg(test)]
 use crate::tests::prelude::*;
 use crate::wchar::{wstr, WString, L};
+use crate::wchar_ext::WExt;
 use crate::wutil::{file_id_for_path, FileId, INVALID_FILE_ID};
+use crate::FLOGF;
 use lru::LruCache;
+#[cfg(feature = "embed-data")]
+use rust_embed::RustEmbed;
 use std::collections::{HashMap, HashSet};
 use std::num::NonZeroUsize;
 use std::time;
@@ -37,6 +43,50 @@ pub struct Autoload {
     cache: Box<AutoloadFileCache>,
 }
 
+#[cfg(feature = "embed-data")]
+#[derive(RustEmbed)]
+#[folder = "share/"]
+pub struct Asset;
+
+#[cfg(feature = "embed-data")]
+pub fn has_asset(cmd: &str) -> bool {
+    Asset::get(cmd).is_some()
+}
+
+#[cfg(not(feature = "embed-data"))]
+pub fn has_asset(_cmd: &str) -> bool {
+    false
+}
+
+#[derive(Clone, Copy, Eq, PartialEq)]
+enum AssetDir {
+    Functions,
+    Completions,
+}
+
+pub enum AutoloadPath {
+    #[cfg(feature = "embed-data")]
+    Embedded(String),
+    Path(WString),
+}
+
+enum AutoloadResult {
+    Path(AutoloadPath),
+    Loaded,
+    Pending,
+    None,
+}
+
+#[cfg(test)]
+impl AutoloadResult {
+    fn is_none(&self) -> bool {
+        matches!(self, AutoloadResult::None)
+    }
+    fn is_some(&self) -> bool {
+        !self.is_none()
+    }
+}
+
 impl Autoload {
     /// Construct an autoloader that loads from the paths given by `env_var_name`.
     pub fn new(env_var_name: &'static wstr) -> Self {
@@ -54,26 +104,65 @@ impl Autoload {
     /// After returning a path, the command is marked in-progress until the caller calls
     /// mark_autoload_finished() with the same command. Note this does not actually execute any
     /// code; it is the caller's responsibility to load the file.
-    pub fn resolve_command(&mut self, cmd: &wstr, env: &dyn Environment) -> Option<WString> {
-        if let Some(var) = env.get(self.env_var_name) {
-            self.resolve_command_impl(cmd, var.as_list())
-        } else {
-            self.resolve_command_impl(cmd, &[])
+    pub fn resolve_command(&mut self, cmd: &wstr, env: &dyn Environment) -> Option<AutoloadPath> {
+        match self.resolve_command_impl(
+            cmd,
+            env.get(self.env_var_name)
+                .as_ref()
+                .map(|var| var.as_list())
+                .unwrap_or_default(),
+        ) {
+            AutoloadResult::Path(path) => {
+                match &path {
+                    #[cfg(feature = "embed-data")]
+                    AutoloadPath::Embedded(_) => {
+                        FLOGF!(autoload, "Embedded: %ls", cmd);
+                    }
+                    AutoloadPath::Path(path) => {
+                        FLOGF!(
+                            autoload,
+                            "Loading %ls from var %ls from path %ls",
+                            cmd,
+                            self.env_var_name,
+                            path
+                        )
+                    }
+                }
+                Some(path)
+            }
+            AutoloadResult::Loaded | AutoloadResult::Pending | AutoloadResult::None => None,
         }
     }
 
     /// Helper to actually perform an autoload.
     /// This is a static function because it executes fish script, and so must be called without
     /// holding any particular locks.
-    pub fn perform_autoload(path: &wstr, parser: &Parser) {
+    pub fn perform_autoload(path: &AutoloadPath, parser: &Parser) {
         // We do the useful part of what exec_subshell does ourselves
         // - we source the file.
         // We don't create a buffer or check ifs or create a read_limit
-
-        let script_source = L!("source ").to_owned() + &escape(path)[..];
         let prev_statuses = parser.get_last_statuses();
         let _put_back = ScopeGuard::new((), |()| parser.set_last_statuses(prev_statuses));
-        parser.eval(&script_source, &IoChain::new());
+        match path {
+            AutoloadPath::Path(p) => {
+                let script_source = L!("source ").to_owned() + &escape(p)[..];
+                parser.eval(&script_source, &IoChain::new());
+            }
+            #[cfg(feature = "embed-data")]
+            AutoloadPath::Embedded(name) => {
+                use crate::common::str2wcstring;
+                use std::sync::Arc;
+                FLOGF!(autoload, "Loading embedded: %ls", name);
+                let emfile = Asset::get(name).expect("Embedded file not found");
+                let src = str2wcstring(&emfile.data);
+                let mut widename = L!("embedded:").to_owned();
+                widename.push_str(name);
+                let ret = parser.eval_file_wstr(src, Arc::new(widename), &IoChain::new(), None);
+                if let Err(msg) = ret {
+                    eprintf!("%ls", msg);
+                }
+            }
+        }
     }
 
     /// Mark that a command previously returned from path_to_autoload is finished autoloading.
@@ -90,7 +179,9 @@ impl Autoload {
     /// Return whether a command could potentially be autoloaded.
     /// This does not actually mark the command as being autoloaded.
     pub fn can_autoload(&mut self, cmd: &wstr) -> bool {
-        self.cache.check(cmd, true /* allow stale */).is_some()
+        self.cache
+            .check(self.env_var_name, cmd, true /* allow stale */)
+            .is_some()
     }
 
     /// Return whether autoloading has been attempted for a command.
@@ -126,10 +217,10 @@ impl Autoload {
 
     /// Like resolve_autoload(), but accepts the paths directly.
     /// This is exposed for testing.
-    fn resolve_command_impl(&mut self, cmd: &wstr, paths: &[WString]) -> Option<WString> {
+    fn resolve_command_impl(&mut self, cmd: &wstr, paths: &[WString]) -> AutoloadResult {
         // Are we currently in the process of autoloading this?
         if self.current_autoloading.contains(cmd) {
-            return None;
+            return AutoloadResult::Pending;
         }
 
         // Check to see if our paths have changed. If so, replace our cache.
@@ -140,20 +231,33 @@ impl Autoload {
         }
 
         // Do we have an entry to load?
-        let file = self.cache.check(cmd, false)?;
+        let Some(file) = self.cache.check(self.env_var_name, cmd, false) else {
+            return AutoloadResult::None;
+        };
+
+        let file_id = match &file {
+            AutoloadableFileInfo::FileInfo(file) => &file.file_id,
+            #[cfg(feature = "embed-data")]
+            AutoloadableFileInfo::EmbeddedPath(_) => &INVALID_FILE_ID,
+        };
 
         // Is this file the same as what we previously autoloaded?
         if let Some(loaded_file) = self.autoloaded_files.get(cmd) {
-            if *loaded_file == file.file_id {
+            if *loaded_file == *file_id {
                 // The file has been autoloaded and is unchanged.
-                return None;
+                return AutoloadResult::Loaded;
             }
         }
 
         // We're going to (tell our caller to) autoload this command.
         self.current_autoloading.insert(cmd.to_owned());
-        self.autoloaded_files.insert(cmd.to_owned(), file.file_id);
-        Some(file.path)
+        self.autoloaded_files
+            .insert(cmd.to_owned(), file_id.clone());
+        AutoloadResult::Path(match file {
+            AutoloadableFileInfo::FileInfo(path) => AutoloadPath::Path(path.path),
+            #[cfg(feature = "embed-data")]
+            AutoloadableFileInfo::EmbeddedPath(path) => AutoloadPath::Embedded(path),
+        })
     }
 }
 
@@ -162,19 +266,28 @@ const AUTOLOAD_STALENESS_INTERVALL: u64 = 15;
 
 /// Represents a file that we might want to autoload.
 #[derive(Clone)]
-struct AutoloadableFile {
+struct FileInfo {
     /// The path to the file.
     path: WString,
     /// The metadata for the file.
     file_id: FileId,
 }
 
+#[derive(Clone)]
+enum AutoloadableFileInfo {
+    /// An on-disk file.
+    FileInfo(FileInfo),
+    /// An embedded file.
+    #[cfg(feature = "embed-data")]
+    EmbeddedPath(String),
+}
+
 // A timestamp is a monotonic point in time.
 type Timestamp = time::Instant;
 type MissesLruCache = LruCache<WString, Timestamp>;
 
 struct KnownFile {
-    file: AutoloadableFile,
+    file: AutoloadableFileInfo,
     last_checked: Timestamp,
 }
 
@@ -222,10 +335,31 @@ impl AutoloadFileCache {
     /// Check if a command `cmd` can be loaded.
     /// If `allow_stale` is true, allow stale entries; otherwise discard them.
     /// This returns an autoloadable file, or none() if there is no such file.
-    fn check(&mut self, cmd: &wstr, allow_stale: bool) -> Option<AutoloadableFile> {
+    fn check(
+        &mut self,
+        env_var_name: &wstr,
+        cmd: &wstr,
+        allow_stale: bool,
+    ) -> Option<AutoloadableFileInfo> {
+        let asset_dir =
+            cfg!(feature = "embed-data")
+                .then_some(())
+                .and_then(|()| match env_var_name {
+                    s if s == "fish_function_path" => Some(AssetDir::Functions),
+                    s if s == "fish_complete_path" => Some(AssetDir::Completions),
+                    _ => None,
+                });
+
         // Check hits.
         if let Some(value) = self.known_files.get(cmd) {
-            if allow_stale || Self::is_fresh(value.last_checked, Self::current_timestamp()) {
+            #[cfg(feature = "embed-data")]
+            let embedded = matches!(value.file, AutoloadableFileInfo::EmbeddedPath(_));
+            #[cfg(not(feature = "embed-data"))]
+            let embedded = false;
+            if allow_stale
+                || embedded
+                || Self::is_fresh(value.last_checked, Self::current_timestamp())
+            {
                 // Re-use this cached hit.
                 return Some(value.file.clone());
             }
@@ -244,7 +378,10 @@ impl AutoloadFileCache {
         }
 
         // We couldn't satisfy this request from the cache. Hit the disk.
-        let file = self.locate_file(cmd);
+        let file = self
+            .locate_file(cmd, asset_dir, false)
+            .or_else(|| self.locate_asset(cmd, asset_dir?))
+            .or_else(|| self.locate_file(cmd, asset_dir, true));
         if let Some(file) = file.as_ref() {
             let old_value = self.known_files.insert(
                 cmd.to_owned(),
@@ -287,7 +424,12 @@ impl AutoloadFileCache {
 
     /// Attempt to find an autoloadable file by searching our path list for a given command.
     /// Return the file, or none() if none.
-    fn locate_file(&self, cmd: &wstr) -> Option<AutoloadableFile> {
+    fn locate_file(
+        &self,
+        cmd: &wstr,
+        asset_dir: Option<AssetDir>,
+        want_generated_completions: bool,
+    ) -> Option<AutoloadableFileInfo> {
         // If the command is empty or starts with NULL (i.e. is empty as a path)
         // we'd try to source the *directory*, which exists.
         // So instead ignore these here.
@@ -300,6 +442,12 @@ impl AutoloadFileCache {
         // Re-use the storage for path.
         let mut path;
         for dir in self.dirs() {
+            if asset_dir == Some(AssetDir::Completions) {
+                // HACK: Ignore generated_completions until we tried the embedded assets
+                if dir.ends_with(L!("/generated_completions")) != want_generated_completions {
+                    continue;
+                }
+            }
             // Construct the path as dir/cmd.fish
             path = dir.to_owned();
             path.push('/');
@@ -309,11 +457,31 @@ impl AutoloadFileCache {
             let file_id = file_id_for_path(&path);
             if file_id != INVALID_FILE_ID {
                 // Found it.
-                return Some(AutoloadableFile { path, file_id });
+                return Some(AutoloadableFileInfo::FileInfo(FileInfo { path, file_id }));
             }
         }
         None
     }
+
+    #[cfg(not(feature = "embed-data"))]
+    fn locate_asset(&self, _cmd: &wstr, _asset_dir: AssetDir) -> Option<AutoloadableFileInfo> {
+        None
+    }
+    #[cfg(feature = "embed-data")]
+    fn locate_asset(&self, cmd: &wstr, asset_dir: AssetDir) -> Option<AutoloadableFileInfo> {
+        // HACK: In cargo tests, this used to never load functions
+        // It will hang for reasons unrelated to this.
+        if cfg!(test) {
+            return None;
+        }
+        let narrow = wcs2string(cmd);
+        let cmdstr = std::str::from_utf8(&narrow).ok()?;
+        let p = match asset_dir {
+            AssetDir::Functions => "functions/".to_owned() + cmdstr + ".fish",
+            AssetDir::Completions => "completions/".to_owned() + cmdstr + ".fish",
+        };
+        has_asset(&p).then_some(AutoloadableFileInfo::EmbeddedPath(p))
+    }
 }
 
 #[test]
@@ -364,36 +532,60 @@ fn test_autoload() {
     autoload.invalidate_cache();
 
     assert!(!autoload.autoload_in_progress(L!("file1")));
-    assert!(autoload.resolve_command_impl(L!("file1"), paths).is_some());
-    assert!(autoload.resolve_command_impl(L!("file1"), paths).is_none());
+    assert!(matches!(
+        autoload.resolve_command_impl(L!("file1"), paths),
+        AutoloadResult::Path(_)
+    ));
+    assert!(matches!(
+        autoload.resolve_command_impl(L!("file1"), paths),
+        AutoloadResult::Pending
+    ));
     assert!(autoload.autoload_in_progress(L!("file1")));
     assert!(autoload.get_autoloaded_commands() == vec![L!("file1")]);
     autoload.mark_autoload_finished(L!("file1"));
     assert!(!autoload.autoload_in_progress(L!("file1")));
     assert!(autoload.get_autoloaded_commands() == vec![L!("file1")]);
 
-    assert!(autoload.resolve_command_impl(L!("file1"), paths).is_none());
+    assert!(matches!(
+        autoload.resolve_command_impl(L!("file1"), paths),
+        AutoloadResult::Loaded
+    ));
     assert!(autoload
         .resolve_command_impl(L!("nothing"), paths)
         .is_none());
     assert!(autoload.resolve_command_impl(L!("file2"), paths).is_some());
-    assert!(autoload.resolve_command_impl(L!("file2"), paths).is_none());
+    assert!(matches!(
+        autoload.resolve_command_impl(L!("file2"), paths),
+        AutoloadResult::Pending
+    ));
     autoload.mark_autoload_finished(L!("file2"));
-    assert!(autoload.resolve_command_impl(L!("file2"), paths).is_none());
+    assert!(matches!(
+        autoload.resolve_command_impl(L!("file2"), paths),
+        AutoloadResult::Loaded
+    ));
     assert!((autoload.get_autoloaded_commands() == vec![L!("file1"), L!("file2")]));
 
     autoload.clear();
     assert!(autoload.resolve_command_impl(L!("file1"), paths).is_some());
     autoload.mark_autoload_finished(L!("file1"));
-    assert!(autoload.resolve_command_impl(L!("file1"), paths).is_none());
+    assert!(matches!(
+        autoload.resolve_command_impl(L!("file1"), paths),
+        AutoloadResult::Loaded
+    ));
     assert!(autoload
         .resolve_command_impl(L!("nothing"), paths)
         .is_none());
     assert!(autoload.resolve_command_impl(L!("file2"), paths).is_some());
-    assert!(autoload.resolve_command_impl(L!("file2"), paths).is_none());
+    assert!(matches!(
+        autoload.resolve_command_impl(L!("file2"), paths),
+        AutoloadResult::Pending
+    ));
     autoload.mark_autoload_finished(L!("file2"));
 
-    assert!(autoload.resolve_command_impl(L!("file1"), paths).is_none());
+    assert!(matches!(
+        autoload.resolve_command_impl(L!("file1"), paths),
+        AutoloadResult::Loaded
+    ));
     touch_file(&sprintf!("%ls/file1.fish", p1));
     autoload.invalidate_cache();
     assert!(autoload.resolve_command_impl(L!("file1"), paths).is_some());
