@@ -5,11 +5,10 @@ use crate::common::{
     escape, escape_string, str2wcstring, valid_var_name, EscapeFlags, EscapeStringStyle,
 };
 use crate::highlight::{colorize, highlight_shell};
-use crate::input::{
-    input_function_get_names, input_mappings, input_terminfo_get_names,
-    input_terminfo_get_sequence, GetSequenceError, InputMappingSet, KeyNameStyle,
+use crate::input::{input_function_get_names, input_mappings, InputMappingSet, KeyNameStyle};
+use crate::key::{
+    self, char_to_symbol, function_key, parse_keys, Key, Modifiers, KEY_NAMES, MAX_FUNCTION_KEY,
 };
-use crate::key::{self, canonicalize_raw_escapes, char_to_symbol, parse_keys, Key, Modifiers};
 use crate::nix::isatty;
 use std::sync::MutexGuard;
 
@@ -26,7 +25,6 @@ struct Options {
     list_modes: bool,
     print_help: bool,
     silent: bool,
-    use_terminfo: bool,
     have_user: bool,
     user: bool,
     have_preset: bool,
@@ -44,7 +42,6 @@ impl Options {
             list_modes: false,
             print_help: false,
             silent: false,
-            use_terminfo: false,
             have_user: false,
             user: false,
             have_preset: false,
@@ -130,25 +127,23 @@ impl BuiltinBind {
                 }
             }
             KeyNameStyle::RawEscapeSequence => {
-                for key in seq {
+                for (i, key) in seq.iter().enumerate() {
                     if key.modifiers == Modifiers::ALT {
-                        out.push_utfstr(&char_to_symbol('\x1b'));
-                        out.push_utfstr(&char_to_symbol(if key.codepoint == key::Escape {
-                            '\x1b'
-                        } else {
-                            key.codepoint
-                        }));
+                        out.push_utfstr(&char_to_symbol('\x1b', i == 0));
+                        out.push_utfstr(&char_to_symbol(
+                            if key.codepoint == key::Escape {
+                                '\x1b'
+                            } else {
+                                key.codepoint
+                            },
+                            false,
+                        ));
                     } else {
                         assert!(key.modifiers.is_none());
-                        out.push_utfstr(&char_to_symbol(key.codepoint));
+                        out.push_utfstr(&char_to_symbol(key.codepoint, i == 0));
                     }
                 }
             }
-            KeyNameStyle::Terminfo(tname) => {
-                // Note that we show -k here because we have an input key name.
-                out.push_str("-k ");
-                out.push_utfstr(&tname);
-            }
         }
 
         // Now show the list of commands.
@@ -203,13 +198,18 @@ impl BuiltinBind {
         }
     }
 
-    /// Print terminfo key binding names to string buffer used for standard output.
-    ///
-    /// \param all if set, all terminfo key binding names will be printed. If not set, only ones that
-    /// are defined for this terminal are printed.
-    fn key_names(&self, all: bool, streams: &mut IoStreams) {
-        let names = input_terminfo_get_names(!all);
-        for name in names {
+    /// Print all named keys to the string buffer used for standard output.
+    fn key_names(&self, streams: &mut IoStreams) {
+        let function_keys: Vec<WString> = (1..=MAX_FUNCTION_KEY)
+            .map(|i| WString::from(Key::from_raw(function_key(i))))
+            .collect();
+        let mut keys: Vec<&wstr> = KEY_NAMES
+            .iter()
+            .map(|(_encoding, name)| *name)
+            .chain(function_keys.iter().map(|s| s.as_utfstr()))
+            .collect();
+        keys.sort_unstable();
+        for name in keys {
             streams.out.appendln(name);
         }
     }
@@ -222,31 +222,6 @@ impl BuiltinBind {
         }
     }
 
-    /// Wraps input_terminfo_get_sequence(), appending the correct error messages as needed.
-    fn get_terminfo_sequence(&self, seq: &wstr, streams: &mut IoStreams) -> Option<WString> {
-        match input_terminfo_get_sequence(seq) {
-            Ok(tseq) => Some(tseq),
-            Err(err) if !self.opts.silent => {
-                let eseq =
-                    escape_string(seq, EscapeStringStyle::Script(EscapeFlags::NO_PRINTABLES));
-                match err {
-                    GetSequenceError::NotFound => streams.err.append(wgettext_fmt!(
-                        "%ls: No key with name '%ls' found\n",
-                        "bind",
-                        eseq
-                    )),
-                    GetSequenceError::NoSeq => streams.err.append(wgettext_fmt!(
-                        "%ls: Key with name '%ls' does not have any mapping\n",
-                        "bind",
-                        eseq
-                    )),
-                };
-                None
-            }
-            Err(_) => None,
-        }
-    }
-
     /// Add specified key binding.
     fn add(
         &mut self,
@@ -262,9 +237,7 @@ impl BuiltinBind {
         let Some(key_seq) = self.compute_seq(streams, seq) else {
             return true;
         };
-        let key_name_style = if self.opts.use_terminfo {
-            KeyNameStyle::Terminfo(seq.to_owned())
-        } else if is_raw_escape_sequence {
+        let key_name_style = if is_raw_escape_sequence {
             KeyNameStyle::RawEscapeSequence
         } else {
             KeyNameStyle::Plain
@@ -275,21 +248,11 @@ impl BuiltinBind {
     }
 
     fn compute_seq(&self, streams: &mut IoStreams, seq: &wstr) -> Option<Vec<Key>> {
-        if self.opts.use_terminfo {
-            let Some(tinfo_seq) = self.get_terminfo_sequence(seq, streams) else {
-                // get_terminfo_sequence already printed the error.
-                return None;
-            };
-            Some(canonicalize_raw_escapes(
-                tinfo_seq.chars().map(Key::from_single_char).collect(),
-            ))
-        } else {
-            match parse_keys(seq) {
-                Ok(keys) => Some(keys),
-                Err(err) => {
-                    streams.err.append(sprintf!("bind: %s\n", err));
-                    None
-                }
+        match parse_keys(seq) {
+            Ok(keys) => Some(keys),
+            Err(err) => {
+                streams.err.append(sprintf!("bind: %s\n", err));
+                None
             }
         }
     }
@@ -385,13 +348,7 @@ impl BuiltinBind {
                     EscapeStringStyle::Script(EscapeFlags::NO_PRINTABLES),
                 );
                 if !self.opts.silent {
-                    if self.opts.use_terminfo {
-                        streams.err.append(wgettext_fmt!(
-                            "%ls: No binding found for key '%ls'\n",
-                            cmd,
-                            eseq
-                        ));
-                    } else if seq.len() == 1 {
+                    if seq.len() == 1 {
                         streams.err.append(wgettext_fmt!(
                             "%ls: No binding found for key '%ls'\n",
                             cmd,
@@ -448,9 +405,9 @@ fn parse_cmd_opts(
     argv: &mut [&wstr],
     parser: &Parser,
     streams: &mut IoStreams,
-) -> Option<i32> {
+) -> BuiltinResult {
     let cmd = argv[0];
-    let short_options = L!(":aehkKfM:Lm:s");
+    let short_options = L!("aehkKfM:Lm:s");
     const long_options: &[WOption] = &[
         wopt(L!("all"), NoArgument, 'a'),
         wopt(L!("erase"), NoArgument, 'e'),
@@ -473,11 +430,17 @@ fn parse_cmd_opts(
             'e' => opts.mode = BIND_ERASE,
             'f' => opts.mode = BIND_FUNCTION_NAMES,
             'h' => opts.print_help = true,
-            'k' => opts.use_terminfo = true,
+            'k' => {
+                streams.err.append(wgettext_fmt!(
+                    "%ls: the -k/--key syntax is no longer supported. See `bind --help` and `bind --key-names`\n",
+                    cmd,
+                ));
+                return Err(STATUS_INVALID_ARGS);
+            }
             'K' => opts.mode = BIND_KEY_NAMES,
             'L' => {
                 opts.list_modes = true;
-                return STATUS_CMD_OK;
+                return Ok(SUCCESS);
             }
             'M' => {
                 if !valid_var_name(w.woptarg.unwrap()) {
@@ -486,7 +449,7 @@ fn parse_cmd_opts(
                         cmd,
                         w.woptarg.unwrap()
                     ));
-                    return STATUS_INVALID_ARGS;
+                    return Err(STATUS_INVALID_ARGS);
                 }
                 opts.bind_mode = w.woptarg.unwrap().to_owned();
                 opts.bind_mode_given = true;
@@ -497,7 +460,7 @@ fn parse_cmd_opts(
                     streams
                         .err
                         .append(wgettext_fmt!(BUILTIN_ERR_BIND_MODE, cmd, new_mode));
-                    return STATUS_INVALID_ARGS;
+                    return Err(STATUS_INVALID_ARGS);
                 }
                 opts.sets_bind_mode = Some(new_mode.to_owned());
             }
@@ -512,11 +475,15 @@ fn parse_cmd_opts(
             }
             ':' => {
                 builtin_missing_argument(parser, streams, cmd, argv[w.wopt_index - 1], true);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
+            }
+            ';' => {
+                builtin_unexpected_argument(parser, streams, cmd, argv[w.wopt_index - 1], true);
+                return Err(STATUS_INVALID_ARGS);
             }
             '?' => {
                 builtin_unknown_option(parser, streams, cmd, argv[w.wopt_index - 1], true);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             }
             _ => {
                 panic!("unexpected retval from WGetopter")
@@ -524,7 +491,7 @@ fn parse_cmd_opts(
         }
     }
     *optind = w.wopt_index;
-    return STATUS_CMD_OK;
+    return Ok(SUCCESS);
 }
 
 impl BuiltinBind {
@@ -534,22 +501,19 @@ impl BuiltinBind {
         parser: &Parser,
         streams: &mut IoStreams,
         argv: &mut [&wstr],
-    ) -> Option<c_int> {
+    ) -> BuiltinResult {
         let cmd = argv[0];
         let mut optind = 0;
-        let retval = parse_cmd_opts(&mut self.opts, &mut optind, argv, parser, streams);
-        if retval != STATUS_CMD_OK {
-            return retval;
-        }
+        parse_cmd_opts(&mut self.opts, &mut optind, argv, parser, streams)?;
 
         if self.opts.list_modes {
             self.list_modes(streams);
-            return STATUS_CMD_OK;
+            return Ok(SUCCESS);
         }
 
         if self.opts.print_help {
             builtin_print_help(parser, streams, cmd);
-            return STATUS_CMD_OK;
+            return Ok(SUCCESS);
         }
 
         // Default to user mode
@@ -567,7 +531,7 @@ impl BuiltinBind {
                         true, /* user */
                         streams,
                     ) {
-                        return STATUS_CMD_ERROR;
+                        return Err(STATUS_CMD_ERROR);
                     }
                 }
                 if self.opts.preset {
@@ -577,28 +541,28 @@ impl BuiltinBind {
                         false, /* user */
                         streams,
                     ) {
-                        return STATUS_CMD_ERROR;
+                        return Err(STATUS_CMD_ERROR);
                     }
                 }
             }
             BIND_INSERT => {
                 if self.insert(optind, argv, parser, streams) {
-                    return STATUS_CMD_ERROR;
+                    return Err(STATUS_CMD_ERROR);
                 }
             }
-            BIND_KEY_NAMES => self.key_names(self.opts.all, streams),
+            BIND_KEY_NAMES => self.key_names(streams),
             BIND_FUNCTION_NAMES => self.function_names(streams),
             _ => {
                 streams
                     .err
                     .append(wgettext_fmt!("%ls: Invalid state\n", cmd));
-                return STATUS_CMD_ERROR;
+                return Err(STATUS_CMD_ERROR);
             }
         }
-        STATUS_CMD_OK
+        Ok(SUCCESS)
     }
 }
 
-pub fn bind(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Option<c_int> {
+pub fn bind(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
     BuiltinBind::new().bind(parser, streams, args)
 }
