@@ -1,25 +1,23 @@
 use super::prelude::*;
 use crate::event;
 
-pub fn emit(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Option<c_int> {
-    let cmd = argv[0];
-
-    let opts = match HelpOnlyCmdOpts::parse(argv, parser, streams) {
-        Ok(opts) => opts,
-        Err(err @ Some(_)) if err != STATUS_CMD_OK => return err,
-        Err(err) => panic!("Illogical exit code from parse_options(): {err:?}"),
+pub fn emit(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> BuiltinResult {
+    let Some(&cmd) = argv.get(0) else {
+        return Err(STATUS_INVALID_ARGS);
     };
 
+    let opts = HelpOnlyCmdOpts::parse(argv, parser, streams)?;
+
     if opts.print_help {
         builtin_print_help(parser, streams, cmd);
-        return STATUS_CMD_OK;
+        return Ok(SUCCESS);
     }
 
     let Some(event_name) = argv.get(opts.optind) else {
         streams
             .err
             .append(sprintf!(L!("%ls: expected event name\n"), cmd));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     };
 
     event::fire_generic(
@@ -31,5 +29,5 @@ pub fn emit(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
             .collect(),
     );
 
-    STATUS_CMD_OK
+    Ok(SUCCESS)
 }
