@@ -154,6 +154,9 @@ struct Options<'args> {
 
     no_ext_valid: bool,
     no_ext: bool,
+
+    all_valid: bool,
+    all: bool,
 }
 
 #[inline]
@@ -175,7 +178,7 @@ fn path_out(streams: &mut IoStreams, opts: &Options<'_>, s: impl AsRef<wstr>) {
 
 fn construct_short_opts(opts: &Options) -> WString {
     // All commands accept -z, -Z and -q
-    let mut short_opts = WString::from(":zZq");
+    let mut short_opts = WString::from("zZq");
     if opts.perms_valid {
         short_opts += L!("p:");
         short_opts += L!("rwx");
@@ -201,14 +204,13 @@ fn construct_short_opts(opts: &Options) -> WString {
     if opts.no_ext_valid {
         short_opts.push('E');
     }
-
     short_opts
 }
 
 /// Note that several long flags share the same short flag. That is okay. The caller is expected
 /// to indicate that a max of one of the long flags sharing a short flag is valid.
 /// Remember: adjust the completions in share/completions/ when options change
-const LONG_OPTIONS: [WOption<'static>; 11] = [
+const LONG_OPTIONS: [WOption<'static>; 12] = [
     wopt(L!("quiet"), NoArgument, 'q'),
     wopt(L!("null-in"), NoArgument, 'z'),
     wopt(L!("null-out"), NoArgument, 'Z'),
@@ -220,6 +222,7 @@ const LONG_OPTIONS: [WOption<'static>; 11] = [
     wopt(L!("unique"), NoArgument, 'u'),
     wopt(L!("key"), RequiredArgument, NON_OPTION_CHAR),
     wopt(L!("no-extension"), NoArgument, 'E'),
+    wopt(L!("all"), NoArgument, '\x02'),
 ];
 
 fn parse_opts<'args>(
@@ -229,7 +232,7 @@ fn parse_opts<'args>(
     args: &mut [&'args wstr],
     parser: &Parser,
     streams: &mut IoStreams,
-) -> Option<c_int> {
+) -> BuiltinResult {
     let cmd = args[0];
     let mut args_read = Vec::with_capacity(args.len());
     args_read.extend_from_slice(args);
@@ -242,11 +245,22 @@ fn parse_opts<'args>(
             ':' => {
                 streams.err.append(L!("path ")); // clone of string_error
                 builtin_missing_argument(parser, streams, cmd, args_read[w.wopt_index - 1], false);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
+            }
+            ';' => {
+                streams.err.append(L!("path ")); // clone of string_error
+                builtin_unexpected_argument(
+                    parser,
+                    streams,
+                    cmd,
+                    args_read[w.wopt_index - 1],
+                    false,
+                );
+                return Err(STATUS_INVALID_ARGS);
             }
             '?' => {
                 path_unknown_option(parser, streams, cmd, args_read[w.wopt_index - 1]);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             }
             'q' => {
                 opts.quiet = true;
@@ -270,7 +284,7 @@ fn parse_opts<'args>(
                 for t in types_args {
                     let Ok(r#type) = t.try_into() else {
                         path_error!(streams, "%ls: Invalid type '%ls'\n", "path", t);
-                        return STATUS_INVALID_ARGS;
+                        return Err(STATUS_INVALID_ARGS);
                     };
                     *types |= r#type;
                 }
@@ -282,7 +296,7 @@ fn parse_opts<'args>(
                 for p in perms_args {
                     let Ok(perm) = p.try_into() else {
                         path_error!(streams, "%ls: Invalid permission '%ls'\n", "path", p);
-                        return STATUS_INVALID_ARGS;
+                        return Err(STATUS_INVALID_ARGS);
                     };
                     *perms |= perm;
                 }
@@ -339,9 +353,14 @@ fn parse_opts<'args>(
                 opts.key = w.woptarg;
                 continue;
             }
+
+            '\x02' if opts.all_valid => {
+                opts.all = true;
+                continue;
+            }
             _ => {
                 path_unknown_option(parser, streams, cmd, args_read[w.wopt_index - 1]);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             }
         }
     }
@@ -357,17 +376,17 @@ fn parse_opts<'args>(
 
         if opts.arg1.is_none() && n_req_args == 1 {
             path_error!(streams, BUILTIN_ERR_ARG_COUNT0, cmd);
-            return STATUS_INVALID_ARGS;
+            return Err(STATUS_INVALID_ARGS);
         }
     }
 
     // At this point we should not have optional args and be reading args from stdin.
     if streams.stdin_is_directly_redirected && args.len() > *optind {
         path_error!(streams, BUILTIN_ERR_TOO_MANY_ARGUMENTS, cmd);
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
-    STATUS_CMD_OK
+    Ok(SUCCESS)
 }
 
 fn path_transform(
@@ -376,14 +395,12 @@ fn path_transform(
     args: &mut [&wstr],
     func: impl Fn(&wstr) -> WString,
     custom_opts: impl Fn(&mut Options),
-) -> Option<c_int> {
+) -> BuiltinResult {
     let mut opts = Options::default();
     custom_opts(&mut opts);
     let mut optind = 0;
-    let retval = parse_opts(&mut opts, &mut optind, 0, args, parser, streams);
-    if retval != STATUS_CMD_OK {
-        return retval;
-    }
+
+    parse_opts(&mut opts, &mut optind, 0, args, parser, streams)?;
 
     let mut n_transformed = 0;
     let arguments = arguments(args, &mut optind, streams).with_split_behavior(match opts.null_in {
@@ -406,20 +423,20 @@ fn path_transform(
             // Return okay if path wasn't already in this form
             // TODO: Is that correct?
             if opts.quiet {
-                return STATUS_CMD_OK;
+                return Ok(SUCCESS);
             };
         }
         path_out(streams, &opts, transformed);
     }
 
     if n_transformed > 0 {
-        STATUS_CMD_OK
+        Ok(SUCCESS)
     } else {
-        STATUS_CMD_ERROR
+        Err(STATUS_CMD_ERROR)
     }
 }
 
-fn path_basename(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Option<c_int> {
+fn path_basename(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
     path_transform(
         parser,
         streams,
@@ -431,7 +448,7 @@ fn path_basename(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -
     )
 }
 
-fn path_dirname(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Option<c_int> {
+fn path_dirname(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
     path_transform(parser, streams, args, |s| wdirname(s).to_owned(), |_| {})
 }
 
@@ -443,18 +460,16 @@ fn normalize_help(path: &wstr) -> WString {
     np
 }
 
-fn path_normalize(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Option<c_int> {
+fn path_normalize(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
     path_transform(parser, streams, args, normalize_help, |_| {})
 }
 
-fn path_mtime(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Option<c_int> {
+fn path_mtime(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
     let mut opts = Options::default();
     opts.relative_valid = true;
     let mut optind = 0;
-    let retval = parse_opts(&mut opts, &mut optind, 0, args, parser, streams);
-    if retval != STATUS_CMD_OK {
-        return retval;
-    }
+
+    parse_opts(&mut opts, &mut optind, 0, args, parser, streams)?;
 
     let mut n_transformed = 0;
 
@@ -472,7 +487,7 @@ fn path_mtime(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> O
 
         if ret != INVALID_FILE_ID {
             if opts.quiet {
-                return STATUS_CMD_OK;
+                return Ok(SUCCESS);
             }
             n_transformed += 1;
             if !opts.relative {
@@ -487,9 +502,9 @@ fn path_mtime(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> O
     }
 
     if n_transformed > 0 {
-        STATUS_CMD_OK
+        Ok(SUCCESS)
     } else {
-        STATUS_CMD_ERROR
+        Err(STATUS_CMD_ERROR)
     }
 }
 
@@ -531,13 +546,11 @@ fn test_find_extension() {
     }
 }
 
-fn path_extension(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Option<c_int> {
+fn path_extension(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
     let mut opts = Options::default();
     let mut optind = 0;
-    let retval = parse_opts(&mut opts, &mut optind, 0, args, parser, streams);
-    if retval != STATUS_CMD_OK {
-        return retval;
-    }
+
+    parse_opts(&mut opts, &mut optind, 0, args, parser, streams)?;
 
     let mut n_transformed = 0;
     let arguments = arguments(args, &mut optind, streams).with_split_behavior(match opts.null_in {
@@ -555,16 +568,16 @@ fn path_extension(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
 
         let ext = arg.slice_from(pos);
         if opts.quiet && !ext.is_empty() {
-            return STATUS_CMD_OK;
+            return Ok(SUCCESS);
         }
         path_out(streams, &opts, ext);
         n_transformed += 1;
     }
 
     if n_transformed > 0 {
-        STATUS_CMD_OK
+        Ok(SUCCESS)
     } else {
-        STATUS_CMD_ERROR
+        Err(STATUS_CMD_ERROR)
     }
 }
 
@@ -572,13 +585,11 @@ fn path_change_extension(
     parser: &Parser,
     streams: &mut IoStreams,
     args: &mut [&wstr],
-) -> Option<c_int> {
+) -> BuiltinResult {
     let mut opts = Options::default();
     let mut optind = 0;
-    let retval = parse_opts(&mut opts, &mut optind, 1, args, parser, streams);
-    if retval != STATUS_CMD_OK {
-        return retval;
-    }
+
+    parse_opts(&mut opts, &mut optind, 1, args, parser, streams)?;
 
     let mut n_transformed = 0usize;
     let arguments = arguments(args, &mut optind, streams).with_split_behavior(match opts.null_in {
@@ -611,19 +622,17 @@ fn path_change_extension(
     }
 
     if n_transformed > 0 {
-        STATUS_CMD_OK
+        Ok(SUCCESS)
     } else {
-        STATUS_CMD_ERROR
+        Err(STATUS_CMD_ERROR)
     }
 }
 
-fn path_resolve(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Option<c_int> {
+fn path_resolve(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
     let mut opts = Options::default();
     let mut optind = 0;
-    let retval = parse_opts(&mut opts, &mut optind, 0, args, parser, streams);
-    if retval != STATUS_CMD_OK {
-        return retval;
-    }
+
+    parse_opts(&mut opts, &mut optind, 0, args, parser, streams)?;
 
     let mut n_transformed = 0usize;
     let arguments = arguments(args, &mut optind, streams).with_split_behavior(match opts.null_in {
@@ -669,28 +678,26 @@ fn path_resolve(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) ->
 
         // Return 0 if we found a realpath.
         if opts.quiet {
-            return STATUS_CMD_OK;
+            return Ok(SUCCESS);
         }
         path_out(streams, &opts, real);
         n_transformed += 1;
     }
 
     if n_transformed > 0 {
-        STATUS_CMD_OK
+        Ok(SUCCESS)
     } else {
-        STATUS_CMD_ERROR
+        Err(STATUS_CMD_ERROR)
     }
 }
 
-fn path_sort(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Option<c_int> {
+fn path_sort(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
     let mut opts = Options::default();
     opts.reverse_valid = true;
     opts.unique_valid = true;
     let mut optind = 0;
-    let retval = parse_opts(&mut opts, &mut optind, 0, args, parser, streams);
-    if retval != STATUS_CMD_OK {
-        return retval;
-    }
+
+    parse_opts(&mut opts, &mut optind, 0, args, parser, streams)?;
 
     let keyfunc: fn(&wstr) -> &wstr = match &opts.key {
         Some(k) if k == "basename" => wbasename,
@@ -703,7 +710,7 @@ fn path_sort(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Op
         None => wbasename,
         Some(k) => {
             path_error!(streams, "%ls: Invalid sort key '%ls'\n", args[0], k);
-            return STATUS_INVALID_ARGS;
+            return Err(STATUS_INVALID_ARGS);
         }
     };
 
@@ -744,7 +751,7 @@ fn path_sort(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Op
     }
 
     /* TODO: Return true only if already sorted? */
-    STATUS_CMD_OK
+    Ok(SUCCESS)
 }
 
 fn filter_path(opts: &Options, path: &wstr, uid: Option<u32>, gid: Option<u32>) -> bool {
@@ -853,16 +860,15 @@ fn path_filter_maybe_is(
     streams: &mut IoStreams,
     args: &mut [&wstr],
     is_is: bool,
-) -> Option<c_int> {
+) -> BuiltinResult {
     let mut opts = Options::default();
     opts.types_valid = true;
     opts.perms_valid = true;
     opts.invert_valid = true;
+    opts.all_valid = true;
     let mut optind = 0;
-    let retval = parse_opts(&mut opts, &mut optind, 0, args, parser, streams);
-    if retval != STATUS_CMD_OK {
-        return retval;
-    }
+
+    parse_opts(&mut opts, &mut optind, 0, args, parser, streams)?;
 
     // If we have been invoked as "path is", which is "path filter -q".
     if is_is {
@@ -887,18 +893,35 @@ fn path_filter_maybe_is(
         None
     };
 
-    for (arg, _) in arguments.filter(|(f, _)| {
-        (opts.perms.is_none() && opts.types.is_none())
-            || (filter_path(&opts, f, uid, gid) != opts.invert)
-    }) {
+    // Collect arguments into a Vec so we can use .len()
+    let arguments_vec: Vec<_> = arguments.collect();
+
+    for (arg, _) in arguments_vec
+        .iter()
+        .filter(|&(f, _)| {
+            (opts.perms.is_none() && opts.types.is_none())
+                || (filter_path(&opts, f, uid, gid) != opts.invert)
+        })
+        .cloned()
+    {
         // If we don't have filters, check if it exists.
         if opts.perms.is_none() && opts.types.is_none() {
             let ok = waccess(&arg, F_OK) == 0;
             if ok == opts.invert {
+                // For --all, fail early if any path does not match the filter.
+                if opts.all {
+                    return Err(STATUS_CMD_ERROR);
+                }
                 continue;
             }
         }
 
+        n_transformed += 1;
+
+        if opts.all {
+            // For --all, do not output paths, just check all must match.
+            continue;
+        }
         // We *know* this is a filename,
         // and so if it starts with a `-` we *know* it is relative
         // to $PWD. So we can add `./`.
@@ -909,43 +932,48 @@ fn path_filter_maybe_is(
         } else {
             path_out(streams, &opts, arg);
         }
-        n_transformed += 1;
         if opts.quiet {
-            return STATUS_CMD_OK;
+            return Ok(SUCCESS);
         };
     }
 
+    if opts.all && n_transformed != arguments_vec.len() {
+        // We have a filter and some paths didn't match.
+        // Return Err if we have a filter and some paths didn't match.
+        return Err(STATUS_CMD_ERROR);
+    }
     if n_transformed > 0 {
-        STATUS_CMD_OK
+        Ok(SUCCESS)
     } else {
-        STATUS_CMD_ERROR
+        Err(STATUS_CMD_ERROR)
     }
 }
 
-fn path_filter(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Option<c_int> {
+fn path_filter(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
     path_filter_maybe_is(parser, streams, args, false)
 }
 
-fn path_is(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Option<c_int> {
+fn path_is(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
     path_filter_maybe_is(parser, streams, args, true)
 }
 
 /// The path builtin, for handling paths.
-pub fn path(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Option<c_int> {
-    let cmd = args[0];
+pub fn path(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
+    let Some(&cmd) = args.get(0) else {
+        return Err(STATUS_INVALID_ARGS);
+    };
     let argc = args.len();
-
     if argc <= 1 {
         streams
             .err
             .append(wgettext_fmt!(BUILTIN_ERR_MISSING_SUBCMD, cmd));
         builtin_print_error_trailer(parser, streams.err, cmd);
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     if args[1] == "-h" || args[1] == "--help" {
         builtin_print_help(parser, streams, cmd);
-        return STATUS_CMD_OK;
+        return Ok(SUCCESS);
     }
 
     let subcmd_name = args[1];
@@ -966,14 +994,14 @@ pub fn path(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Opt
                 .err
                 .append(wgettext_fmt!(BUILTIN_ERR_INVALID_SUBCMD, cmd, subcmd_name));
             builtin_print_error_trailer(parser, streams.err, cmd);
-            return STATUS_INVALID_ARGS;
+            return Err(STATUS_INVALID_ARGS);
         }
     };
 
     if argc >= 3 && (args[2] == "-h" || args[2] == "--help") {
         // Unlike string, we don't have separate docs (yet)
         builtin_print_help(parser, streams, cmd);
-        return STATUS_CMD_OK;
+        return Ok(SUCCESS);
     }
     let args = &mut args[1..];
     return subcmd(parser, streams, args);
