@@ -79,7 +79,7 @@ struct builtin_printf_state_t<'a, 'b> {
     streams: &'a mut IoStreams<'b>,
 
     // The status of the operation.
-    exit_code: c_int,
+    exit_code: BuiltinResult,
 
     // Whether we should stop outputting. This gets set in the case of an error, and also with the
     // \c escape.
@@ -588,7 +588,7 @@ impl<'a, 'b> builtin_printf_state_t<'a, 'b> {
 
         // We set the exit code to error, because one occurred,
         // but we don't do an early exit so we still print what we can.
-        self.exit_code = STATUS_CMD_ERROR.unwrap();
+        self.exit_code = Err(STATUS_CMD_ERROR);
     }
 
     fn fatal_error<Str: AsRef<wstr>>(&mut self, errstr: Str) {
@@ -610,7 +610,7 @@ impl<'a, 'b> builtin_printf_state_t<'a, 'b> {
             self.streams.err.push('\n');
         }
 
-        self.exit_code = STATUS_CMD_ERROR.unwrap();
+        self.exit_code = Err(STATUS_CMD_ERROR);
         self.early_exit = true;
     }
 
@@ -767,19 +767,19 @@ impl<'a, 'b> builtin_printf_state_t<'a, 'b> {
 }
 
 /// The printf builtin.
-pub fn printf(_parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Option<c_int> {
+pub fn printf(_parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> BuiltinResult {
     let mut argc = argv.len();
 
     // Rebind argv as immutable slice (can't rearrange its elements), skipping the command name.
     let mut argv: &[&wstr] = &argv[1..];
     argc -= 1;
     if argc < 1 {
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     let mut state = builtin_printf_state_t {
         streams,
-        exit_code: STATUS_CMD_OK.unwrap(),
+        exit_code: Ok(SUCCESS),
         early_exit: false,
         buff: WString::new(),
         locale: get_numeric_locale(),
@@ -799,5 +799,5 @@ pub fn printf(_parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) ->
             break;
         }
     }
-    return Some(state.exit_code);
+    return state.exit_code;
 }
