@@ -3,7 +3,6 @@
 use super::prelude::*;
 use crate::common::escape;
 use crate::common::read_blocked;
-use crate::common::scoped_push_replacer;
 use crate::common::str2wcstring;
 use crate::common::unescape_string;
 use crate::common::valid_var_name;
@@ -13,23 +12,35 @@ use crate::env::Environment;
 use crate::env::READ_BYTE_LIMIT;
 use crate::env::{EnvVar, EnvVarFlags};
 use crate::input_common::decode_input_byte;
-use crate::input_common::terminal_protocols_disable_ifn;
 use crate::input_common::DecodeState;
+use crate::input_common::InvalidPolicy;
 use crate::nix::isatty;
 use crate::reader::commandline_set_buffer;
+use crate::reader::reader_save_screen_state;
 use crate::reader::ReaderConfig;
 use crate::reader::{reader_pop, reader_push, reader_readline};
+use crate::tokenizer::Tok;
 use crate::tokenizer::Tokenizer;
 use crate::tokenizer::TOK_ACCEPT_UNFINISHED;
+use crate::tokenizer::TOK_ARGUMENT_LIST;
+use crate::tty_handoff::TtyHandoff;
 use crate::wcstringutil::split_about;
 use crate::wcstringutil::split_string_tok;
 use crate::wutil;
 use crate::wutil::encoding::zero_mbstate;
 use crate::wutil::perror;
 use libc::SEEK_CUR;
+use std::num::NonZeroUsize;
 use std::os::fd::RawFd;
 use std::sync::atomic::Ordering;
 
+#[derive(Clone, Copy, Eq, PartialEq)]
+pub(crate) enum TokenOutputMode {
+    Expanded,
+    Raw,
+    Unescaped,
+}
+
 #[derive(Default)]
 struct Options {
     print_help: bool,
@@ -37,17 +48,17 @@ struct Options {
     prompt: Option<WString>,
     prompt_str: Option<WString>,
     right_prompt: WString,
-    commandline: WString,
+    commandline: Option<WString>,
     // If a delimiter was given. Used to distinguish between the default
     // empty string and a given empty delimiter.
     delimiter: Option<WString>,
-    tokenize: bool,
+    token_mode: Option<TokenOutputMode>, // never expanded
     shell: bool,
     array: bool,
     silent: bool,
     split_null: bool,
     to_stdout: bool,
-    nchars: usize,
+    nchars: Option<NonZeroUsize>,
     one_line: bool,
 }
 
@@ -60,7 +71,7 @@ impl Options {
     }
 }
 
-const SHORT_OPTIONS: &wstr = L!(":ac:d:fghiLln:p:sStuxzP:UR:L");
+const SHORT_OPTIONS: &wstr = L!("ac:d:fghiLln:p:sStuxzP:UR:L");
 const LONG_OPTIONS: &[WOption] = &[
     wopt(L!("array"), ArgType::NoArgument, 'a'),
     wopt(L!("command"), ArgType::RequiredArgument, 'c'),
@@ -80,15 +91,24 @@ const LONG_OPTIONS: &[WOption] = &[
     wopt(L!("shell"), ArgType::NoArgument, 'S'),
     wopt(L!("silent"), ArgType::NoArgument, 's'),
     wopt(L!("tokenize"), ArgType::NoArgument, 't'),
+    wopt(L!("tokenize-raw"), ArgType::NoArgument, '\x01'),
     wopt(L!("unexport"), ArgType::NoArgument, 'u'),
     wopt(L!("universal"), ArgType::NoArgument, 'U'),
 ];
 
+fn tokenize_flag(token_mode: TokenOutputMode) -> &'static wstr {
+    match token_mode {
+        TokenOutputMode::Expanded => panic!(),
+        TokenOutputMode::Raw => L!("--tokenize-raw"),
+        TokenOutputMode::Unescaped => L!("--tokenize"),
+    }
+}
+
 fn parse_cmd_opts(
     args: &mut [&wstr],
     parser: &Parser,
     streams: &mut IoStreams,
-) -> Result<(Options, usize), Option<c_int>> {
+) -> Result<(Options, usize), ErrorCode> {
     let cmd = args[0];
     let mut opts = Options::new();
     let mut w = WGetopter::new(SHORT_OPTIONS, LONG_OPTIONS, args);
@@ -98,17 +118,14 @@ fn parse_cmd_opts(
                 opts.array = true;
             }
             'c' => {
-                opts.commandline = w.woptarg.unwrap().to_owned();
+                opts.commandline = Some(w.woptarg.unwrap().to_owned());
             }
             'd' => {
                 opts.delimiter = Some(w.woptarg.unwrap().to_owned());
             }
             'i' => {
                 streams.err.append(wgettext_fmt!(
-                    concat!(
-                        "%ls: usage of -i for --silent is deprecated. Please ",
-                        "use -s or --silent instead.\n"
-                    ),
+                    "%ls: usage of -i for --silent is deprecated. Please use -s or --silent instead.\n",
                     cmd
                 ));
                 return Err(STATUS_INVALID_ARGS);
@@ -130,7 +147,7 @@ fn parse_cmd_opts(
             }
             'n' => {
                 opts.nchars = match fish_wcstoi(w.woptarg.unwrap()) {
-                    Ok(n) if n >= 0 => n.try_into().unwrap(),
+                    Ok(n) if n >= 0 => NonZeroUsize::new(n.try_into().unwrap()),
                     Err(wutil::Error::Overflow) => {
                         streams.err.append(wgettext_fmt!(
                             "%ls: Argument '%ls' is out of range\n",
@@ -166,8 +183,28 @@ fn parse_cmd_opts(
             'S' => {
                 opts.shell = true;
             }
-            't' => {
-                opts.tokenize = true;
+            't' | '\x01' => {
+                let new_mode = match opt {
+                    't' => TokenOutputMode::Unescaped,
+                    '\x01' => TokenOutputMode::Raw,
+                    _ => unreachable!(),
+                };
+                if let Some(old_mode) = opts.token_mode {
+                    if old_mode != new_mode {
+                        streams.err.append(wgettext_fmt!(
+                            BUILTIN_ERR_COMBO2,
+                            cmd,
+                            wgettext_fmt!(
+                                "%s and %s are mutually exclusive",
+                                tokenize_flag(old_mode),
+                                tokenize_flag(new_mode),
+                            )
+                        ));
+                        builtin_print_error_trailer(parser, streams.err, cmd);
+                        return Err(STATUS_INVALID_ARGS);
+                    }
+                }
+                opts.token_mode = Some(new_mode);
             }
             'U' => {
                 opts.place |= EnvMode::UNIVERSAL;
@@ -185,6 +222,10 @@ fn parse_cmd_opts(
                 builtin_missing_argument(parser, streams, cmd, args[w.wopt_index - 1], true);
                 return Err(STATUS_INVALID_ARGS);
             }
+            ';' => {
+                builtin_unexpected_argument(parser, streams, cmd, args[w.wopt_index - 1], true);
+                return Err(STATUS_INVALID_ARGS);
+            }
             '?' => {
                 builtin_unknown_option(parser, streams, cmd, args[w.wopt_index - 1], true);
                 return Err(STATUS_INVALID_ARGS);
@@ -203,15 +244,15 @@ fn parse_cmd_opts(
 fn read_interactive(
     parser: &Parser,
     buff: &mut WString,
-    nchars: usize,
+    nchars: Option<NonZeroUsize>,
     shell: bool,
     silent: bool,
     prompt: &wstr,
     right_prompt: &wstr,
-    commandline: &wstr,
+    commandline: &Option<WString>,
     inputfd: RawFd,
-) -> Option<c_int> {
-    let mut exit_res = STATUS_CMD_OK;
+) -> BuiltinResult {
+    let mut exit_res = Ok(SUCCESS);
 
     // Construct a configuration.
     let mut conf = ReaderConfig::default();
@@ -234,32 +275,34 @@ fn read_interactive(
 
     // Keep in-memory history only.
     reader_push(parser, L!(""), conf);
-    let _modifiable_commandline = scoped_push_replacer(
-        |new_value| std::mem::replace(&mut parser.libdata_mut().readonly_commandline, new_value),
-        false,
-    );
-    commandline_set_buffer(parser, Some(commandline.to_owned()), None);
+    let _modifiable_commandline = parser.scope().readonly_commandline.then(|| {
+        parser.push_scope(|s| {
+            s.readonly_commandline = false;
+        })
+    });
+    if let Some(commandline) = commandline {
+        commandline_set_buffer(parser, Some(commandline.clone()), None);
+    }
 
     let mline = {
-        let _interactive = scoped_push_replacer(
-            |new_value| std::mem::replace(&mut parser.libdata_mut().is_interactive, new_value),
-            true,
-        );
-
+        let _interactive = parser.push_scope(|s| s.is_interactive = true);
+        let mut scoped_handoff = TtyHandoff::new(reader_save_screen_state);
+        scoped_handoff.enable_tty_protocols();
         reader_readline(parser, nchars)
     };
-    terminal_protocols_disable_ifn(false);
     if let Some(line) = mline {
         *buff = line;
-        if nchars > 0 && nchars < buff.len() {
+        if let Some(nchars) = nchars.map(usize::from) {
             // Line may be longer than nchars if a keybinding used `commandline -i`
             // note: we're deliberately throwing away the tail of the commandline.
             // It shouldn't be unread because it was produced with `commandline -i`,
             // not typed.
-            buff.truncate(nchars);
+            if nchars < buff.len() {
+                buff.truncate(nchars);
+            }
         }
     } else {
-        exit_res = STATUS_CMD_ERROR;
+        exit_res = Err(STATUS_CMD_ERROR);
     }
     reader_pop();
     exit_res
@@ -275,8 +318,8 @@ const READ_CHUNK_SIZE: usize = 128;
 /// of chars.
 ///
 /// Returns an exit status.
-fn read_in_chunks(fd: RawFd, buff: &mut WString, split_null: bool, do_seek: bool) -> Option<c_int> {
-    let mut exit_res = STATUS_CMD_OK;
+fn read_in_chunks(fd: RawFd, buff: &mut WString, split_null: bool, do_seek: bool) -> BuiltinResult {
+    let mut exit_res = Ok(SUCCESS);
     let mut narrow_buff = vec![];
     let mut eof = false;
     let mut finished = false;
@@ -314,18 +357,18 @@ fn read_in_chunks(fd: RawFd, buff: &mut WString, split_null: bool, do_seek: bool
                 } == -1
             {
                 perror("lseek");
-                return STATUS_CMD_ERROR;
+                return Err(STATUS_CMD_ERROR);
             }
             finished = true;
         } else if narrow_buff.len() > READ_BYTE_LIMIT.load(Ordering::Relaxed) {
-            exit_res = STATUS_READ_TOO_MUCH;
+            exit_res = Err(STATUS_READ_TOO_MUCH);
             finished = true;
         }
     }
 
     *buff = str2wcstring(&narrow_buff);
     if buff.is_empty() && eof {
-        exit_res = STATUS_CMD_ERROR;
+        exit_res = Err(STATUS_CMD_ERROR);
     }
 
     exit_res
@@ -337,10 +380,10 @@ fn read_in_chunks(fd: RawFd, buff: &mut WString, split_null: bool, do_seek: bool
 fn read_one_char_at_a_time(
     fd: RawFd,
     buff: &mut WString,
-    nchars: usize,
+    nchars: Option<NonZeroUsize>,
     split_null: bool,
-) -> Option<c_int> {
-    let mut exit_res = STATUS_CMD_OK;
+) -> BuiltinResult {
+    let mut exit_res = Ok(SUCCESS);
     let mut nbytes = 0;
 
     let mut unconsumed = vec![];
@@ -361,29 +404,32 @@ fn read_one_char_at_a_time(
             unconsumed.push(b);
             nbytes += 1;
             let mut consumed = 0;
-            match decode_input_byte(buff, &mut state, &unconsumed, &mut consumed) {
+            match decode_input_byte(
+                buff,
+                InvalidPolicy::Passthrough,
+                &mut state,
+                &unconsumed,
+                &mut consumed,
+            ) {
                 DecodeState::Incomplete => continue,
                 DecodeState::Complete => {
                     unconsumed.clear();
                     break Some(buff.as_char_slice().last().unwrap());
                 }
-                DecodeState::Error => {
-                    state = zero_mbstate();
-                    unconsumed.clear();
-                }
+                DecodeState::Error => unreachable!(),
             }
         };
 
         if nbytes > READ_BYTE_LIMIT.load(Ordering::Relaxed) {
             // Historical behavior: do not include the codepoint that made us overflow.
             buff.truncate(chars_read);
-            exit_res = STATUS_READ_TOO_MUCH;
+            exit_res = Err(STATUS_READ_TOO_MUCH);
             break;
         }
         let Some(&res) = res else {
             // EOF
             if buff.is_empty() {
-                exit_res = STATUS_CMD_ERROR;
+                exit_res = Err(STATUS_CMD_ERROR);
             }
             break;
         };
@@ -391,8 +437,10 @@ fn read_one_char_at_a_time(
             buff.pop();
             break;
         }
-        if nchars > 0 && nchars <= buff.len() {
-            break;
+        if let Some(nchars) = nchars.map(usize::from) {
+            if nchars <= buff.len() {
+                break;
+            }
         }
     }
 
@@ -406,7 +454,7 @@ fn validate_read_args(
     argv: &[&wstr],
     parser: &Parser,
     streams: &mut IoStreams,
-) -> Option<c_int> {
+) -> BuiltinResult {
     if opts.prompt.is_some() && opts.prompt_str.is_some() {
         streams.err.append(wgettext_fmt!(
             "%ls: Options %ls and %ls cannot be used together\n",
@@ -415,7 +463,7 @@ fn validate_read_args(
             "-P",
         ));
         builtin_print_error_trailer(parser, streams.err, cmd);
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     if opts.delimiter.is_some() && opts.one_line {
@@ -425,7 +473,7 @@ fn validate_read_args(
             "--delimiter",
             "--line"
         ));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
     if opts.one_line && opts.split_null {
         streams.err.append(wgettext_fmt!(
@@ -434,7 +482,7 @@ fn validate_read_args(
             "-z",
             "--line"
         ));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     if let Some(prompt_str) = opts.prompt_str.as_ref() {
@@ -446,7 +494,7 @@ fn validate_read_args(
     if opts.place.contains(EnvMode::UNEXPORT) && opts.place.contains(EnvMode::EXPORT) {
         streams.err.append(wgettext_fmt!(BUILTIN_ERR_EXPUNEXP, cmd));
         builtin_print_error_trailer(parser, streams.err, cmd);
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     if opts
@@ -458,7 +506,7 @@ fn validate_read_args(
     {
         streams.err.append(wgettext_fmt!(BUILTIN_ERR_GLOCAL, cmd));
         builtin_print_error_trailer(parser, streams.err, cmd);
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     let argc = argv.len();
@@ -466,41 +514,51 @@ fn validate_read_args(
         streams
             .err
             .append(wgettext_fmt!(BUILTIN_ERR_MIN_ARG_COUNT1, cmd, 1, argc));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     if opts.array && argc != 1 {
         streams
             .err
             .append(wgettext_fmt!(BUILTIN_ERR_ARG_COUNT1, cmd, 1, argc));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     if opts.to_stdout && argc > 0 {
         streams
             .err
             .append(wgettext_fmt!(BUILTIN_ERR_MAX_ARG_COUNT1, cmd, 0, argc));
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
-    if opts.tokenize && opts.delimiter.is_some() {
-        streams.err.append(wgettext_fmt!(
-            BUILTIN_ERR_COMBO2_EXCLUSIVE,
-            cmd,
-            "--delimiter",
-            "--tokenize"
-        ));
-        return STATUS_INVALID_ARGS;
+    fn tokenize_flag(token_mode: TokenOutputMode) -> &'static wstr {
+        match token_mode {
+            TokenOutputMode::Expanded => panic!(),
+            TokenOutputMode::Raw => L!("--tokenize-raw"),
+            TokenOutputMode::Unescaped => L!("--tokenize"),
+        }
     }
 
-    if opts.tokenize && opts.one_line {
-        streams.err.append(wgettext_fmt!(
-            BUILTIN_ERR_COMBO2_EXCLUSIVE,
-            cmd,
-            "--line",
-            "--tokenize"
-        ));
-        return STATUS_INVALID_ARGS;
+    if let Some(token_mode) = opts.token_mode {
+        if opts.delimiter.is_some() {
+            streams.err.append(wgettext_fmt!(
+                BUILTIN_ERR_COMBO2_EXCLUSIVE,
+                cmd,
+                "--delimiter",
+                tokenize_flag(token_mode),
+            ));
+            return Err(STATUS_INVALID_ARGS);
+        }
+
+        if opts.one_line {
+            streams.err.append(wgettext_fmt!(
+                BUILTIN_ERR_COMBO2_EXCLUSIVE,
+                cmd,
+                "--line",
+                tokenize_flag(token_mode),
+            ));
+            return Err(STATUS_INVALID_ARGS);
+        }
     }
 
     // Verify all variable names.
@@ -510,7 +568,7 @@ fn validate_read_args(
                 .err
                 .append(wgettext_fmt!(BUILTIN_ERR_VARNAME, cmd, arg));
             builtin_print_error_trailer(parser, streams.err, cmd);
-            return STATUS_INVALID_ARGS;
+            return Err(STATUS_INVALID_ARGS);
         }
         if EnvVar::flags_for(arg).contains(EnvVarFlags::READ_ONLY) {
             streams.err.append(wgettext_fmt!(
@@ -519,22 +577,20 @@ fn validate_read_args(
                 arg
             ));
             builtin_print_error_trailer(parser, streams.err, cmd);
-            return STATUS_INVALID_ARGS;
+            return Err(STATUS_INVALID_ARGS);
         }
     }
 
-    STATUS_CMD_OK
+    Ok(SUCCESS)
 }
 
 /// The read builtin. Reads from stdin and stores the values in environment variables.
-pub fn read(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Option<c_int> {
+pub fn read(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> BuiltinResult {
     let mut buff = WString::new();
-    let mut exit_res;
+    let mut exit_res: BuiltinResult;
+
+    let (mut opts, optind) = parse_cmd_opts(argv, parser, streams)?;
 
-    let (mut opts, optind) = match parse_cmd_opts(argv, parser, streams) {
-        Ok(res) => res,
-        Err(retval) => return retval,
-    };
     let cmd = argv[0];
     let mut argv: &[&wstr] = argv;
     if !opts.to_stdout {
@@ -548,20 +604,17 @@ pub fn read(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
 
     if opts.print_help {
         builtin_print_help(parser, streams, cmd);
-        return STATUS_CMD_OK;
+        return Ok(SUCCESS);
     }
 
-    let retval = validate_read_args(cmd, &mut opts, argv, parser, streams);
-    if retval != STATUS_CMD_OK {
-        return retval;
-    }
+    validate_read_args(cmd, &mut opts, argv, parser, streams)?;
 
     // stdin may have been explicitly closed
     if streams.stdin_fd < 0 {
         streams
             .err
             .append(wgettext_fmt!("%ls: stdin is closed\n", cmd));
-        return STATUS_CMD_ERROR;
+        return Err(STATUS_CMD_ERROR);
     }
 
     if opts.one_line {
@@ -601,7 +654,7 @@ pub fn read(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
                 &opts.commandline,
                 streams.stdin_fd,
             );
-        } else if opts.nchars == 0 && !stream_stdin_is_a_tty &&
+        } else if opts.nchars.is_none() && !stream_stdin_is_a_tty &&
                    // "one_line" is implemented as reading n-times to a new line,
                    // if we're chunking we could get multiple lines so we would have to advance
                    // more than 1 per run through the loop. Let's skip that for now.
@@ -628,7 +681,7 @@ pub fn read(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
                 read_one_char_at_a_time(streams.stdin_fd, &mut buff, opts.nchars, opts.split_null);
         }
 
-        if exit_res != STATUS_CMD_OK {
+        if exit_res.is_err() {
             clear_remaining_vars(&mut var_ptr);
             return exit_res;
         }
@@ -638,18 +691,28 @@ pub fn read(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
             return exit_res;
         }
 
-        if opts.tokenize {
-            let mut tok = Tokenizer::new(&buff, TOK_ACCEPT_UNFINISHED);
+        if let Some(token_mode) = opts.token_mode {
+            let mut tok = Tokenizer::new(&buff, TOK_ACCEPT_UNFINISHED | TOK_ARGUMENT_LIST);
+            let token_text = |tokenizer: &mut Tokenizer<'_>, token: &Tok| -> WString {
+                let mut text = Cow::Borrowed(tokenizer.text_of(token));
+                match token_mode {
+                    TokenOutputMode::Expanded => panic!(),
+                    TokenOutputMode::Raw => (),
+                    TokenOutputMode::Unescaped => {
+                        if let Some(unescaped) =
+                            unescape_string(&text, UnescapeStringStyle::default())
+                        {
+                            text = Cow::Owned(unescaped);
+                        }
+                    }
+                };
+                text.into_owned()
+            };
             if opts.array {
                 // Array mode: assign each token as a separate element of the sole var.
                 let mut tokens = vec![];
                 while let Some(t) = tok.next() {
-                    let text = tok.text_of(&t);
-                    if let Some(out) = unescape_string(text, UnescapeStringStyle::default()) {
-                        tokens.push(out);
-                    } else {
-                        tokens.push(text.to_owned());
-                    }
+                    tokens.push(token_text(&mut tok, &t));
                 }
 
                 parser.set_var_and_fire(argv[var_ptr], opts.place, tokens);
@@ -659,9 +722,7 @@ pub fn read(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
                     let Some(t) = tok.next() else {
                         break;
                     };
-                    let text = tok.text_of(&t);
-                    let out = unescape_string(text, UnescapeStringStyle::default())
-                        .unwrap_or_else(|| text.to_owned());
+                    let out = token_text(&mut tok, &t);
                     parser.set_var_and_fire(argv[var_ptr], opts.place, vec![out]);
                     var_ptr += 1;
                 }
@@ -681,15 +742,15 @@ pub fn read(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
             continue;
         }
 
-        // todo!("don't clone")
-        let delimiter = opts
-            .delimiter
-            .clone()
-            .or_else(|| {
-                let ifs = parser.vars().get_unless_empty(L!("IFS"));
-                ifs.map(|ifs| ifs.as_string())
-            })
-            .unwrap_or_default();
+        let mut ifs_delimiter = WString::new();
+        let delimiter: &wstr = opts.delimiter.as_deref().unwrap_or_else(|| {
+            ifs_delimiter = parser
+                .vars()
+                .get_unless_empty(L!("IFS"))
+                .map(|var| var.as_string())
+                .unwrap_or_default();
+            &ifs_delimiter
+        });
 
         if delimiter.is_empty() {
             // Every character is a separate token with one wrinkle involving non-array mode where
@@ -730,7 +791,7 @@ pub fn read(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
             if opts.delimiter.is_none() {
                 // We're using IFS, so tokenize the buffer using each IFS char. This is for backward
                 // compatibility with old versions of fish.
-                let tokens = split_string_tok(&buff, &delimiter, None)
+                let tokens = split_string_tok(&buff, delimiter, None)
                     .into_iter()
                     .map(|s| s.to_owned())
                     .collect();
@@ -738,7 +799,7 @@ pub fn read(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
                 var_ptr += 1;
             } else {
                 // We're using a delimiter provided by the user so use the `string split` behavior.
-                let splits = split_about(&buff, &delimiter, usize::MAX, false)
+                let splits = split_about(&buff, delimiter, usize::MAX, false)
                     .into_iter()
                     .map(|s| s.to_owned())
                     .collect();
@@ -752,7 +813,7 @@ pub fn read(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
                 // compatibility with old versions of fish.
                 // Note the final variable gets any remaining text.
                 let mut var_vals: Vec<WString> =
-                    split_string_tok(&buff, &delimiter, Some(vars_left(var_ptr)))
+                    split_string_tok(&buff, delimiter, Some(vars_left(var_ptr)))
                         .into_iter()
                         .map(|s| s.to_owned())
                         .collect();
@@ -770,7 +831,7 @@ pub fn read(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Opt
                 // We're using a delimiter provided by the user so use the `string split` behavior.
                 // We're making at most argc - 1 splits so the last variable
                 // is set to the remaining string.
-                let splits = split_about(&buff, &delimiter, argc - 1, false);
+                let splits = split_about(&buff, delimiter, argc - 1, false);
                 assert!(splits.len() <= vars_left(var_ptr));
                 for split in splits {
                     parser.set_var_and_fire(argv[var_ptr], opts.place, vec![split.to_owned()]);
