@@ -32,6 +32,8 @@ use std::sync::atomic::{AtomicI32, AtomicU32, Ordering};
 use std::sync::{Arc, MutexGuard};
 use std::time;
 
+pub const BUILD_DIR: &str = env!("FISH_RESOLVED_BUILD_DIR");
+
 pub const PACKAGE_NAME: &str = env!("CARGO_PKG_NAME");
 
 // Highest legal ASCII value.
@@ -1654,7 +1656,7 @@ pub fn get_executable_path(argv0: impl AsRef<Path>) -> PathBuf {
         // When /proc/self/exe points to a file that was deleted (or overwritten on update!)
         // then linux adds a " (deleted)" suffix.
         // If that's not a valid path, let's remove that awkward suffix.
-        if !path.ends_with(" (deleted)") {
+        if path.as_os_str().as_bytes().ends_with(b" (deleted)") {
             return path;
         }
 
