@@ -10,10 +10,11 @@ use std::{
 };
 
 use crate::{
+    ast::unescape_keyword,
     common::charptr2wcstring,
     reader::{get_quote, is_backslashed},
     util::wcsfilecmp,
-    wutil::sprintf,
+    wutil::{localizable_string, sprintf, LocalizableString},
 };
 use bitflags::bitflags;
 use once_cell::sync::Lazy;
@@ -44,14 +45,14 @@ use crate::{
     parser_keywords::parser_keywords_is_subcommand,
     path::{path_get_path, path_try_get_path},
     tokenizer::{variable_assignment_equals_pos, Tok, TokFlags, TokenType, Tokenizer},
-    wchar::{wstr, WString, L},
+    wchar::prelude::*,
     wchar_ext::WExt,
     wcstringutil::{
         string_fuzzy_match_string, string_prefixes_string, string_prefixes_string_case_insensitive,
         StringFuzzyMatch,
     },
     wildcard::{wildcard_complete, wildcard_has, wildcard_match},
-    wutil::{gettext::wgettext_str, wgettext, wrealpath},
+    wutil::wrealpath,
 };
 
 // Completion description strings, mostly for different types of files, such as sockets, block
@@ -69,18 +70,6 @@ static COMPLETE_VAR_DESC_VAL: Lazy<&wstr> = Lazy::new(|| wgettext!("Variable: %l
 /// Description for abbreviations.
 static ABBR_DESC: Lazy<&wstr> = Lazy::new(|| wgettext!("Abbreviation: %ls"));
 
-/// The special cased translation macro for completions. The empty string needs to be special cased,
-/// since it can occur, and should not be translated. (Gettext returns the version information as
-/// the response).
-#[allow(non_snake_case)]
-fn C_(s: &wstr) -> &'static wstr {
-    if s.is_empty() {
-        L!("")
-    } else {
-        wgettext_str(s)
-    }
-}
-
 #[derive(Clone, Copy, Default, PartialEq, Eq, Debug)]
 pub struct CompletionMode {
     /// If set, skip file completions.
@@ -117,6 +106,8 @@ bitflags! {
         const REPLACES_LINE = 1 << 7;
         /// If replacing the entire token, keep the "foo=" prefix.
         const KEEP_VARIABLE_OVERRIDE_PREFIX = 1 << 8;
+        /// This is a variable name.
+        const VARIABLE_NAME = 1 << 9;
     }
 }
 
@@ -382,7 +373,7 @@ struct CompleteEntryOpt {
     /// Arguments to the option; may be a subshell expression expanded at evaluation time.
     comp: WString,
     /// Description of the completion.
-    desc: WString,
+    desc: LocalizableString,
     /// Conditions under which to use the option, expanded and evaluated at completion time.
     conditions: Vec<WString>,
     /// Type of the option: `ArgsOnly`, `Short`, `SingleLong`, or `DoubleLong`.
@@ -394,10 +385,6 @@ struct CompleteEntryOpt {
 }
 
 impl CompleteEntryOpt {
-    pub fn localized_desc(&self) -> &'static wstr {
-        C_(&self.desc)
-    }
-
     pub fn expected_dash_count(&self) -> usize {
         match self.typ {
             CompleteOptionType::ArgsOnly => 0,
@@ -669,7 +656,12 @@ impl<'ctx> Completer<'ctx> {
         if is_autosuggest {
             let prefixed_supercommand_count = tokens
                 .iter()
-                .take_while(|token| parser_keywords_is_subcommand(token.get_source(&cmdline)))
+                .take_while(|token| {
+                    parser_keywords_is_subcommand(&unescape_keyword(
+                        token.type_,
+                        token.get_source(&cmdline),
+                    ))
+                })
                 .count();
             tokens.drain(..prefixed_supercommand_count);
         }
@@ -753,8 +745,8 @@ impl<'ctx> Completer<'ctx> {
             }
             // Complete command filename.
             let current_token = current_token.to_owned();
-            self.complete_cmd(current_token.clone());
-            self.complete_abbr(current_token);
+            self.complete_abbr(current_token.clone());
+            self.complete_cmd(current_token);
             return;
         }
         // See whether we are in an argument, in a redirection or in the whitespace in between.
@@ -884,7 +876,8 @@ impl<'ctx> Completer<'ctx> {
             // Compute new value and reinsert it.
             let test_res = exec_subshell(
                 condition, parser, None, false, /* don't apply exit status */
-            ) == 0;
+            )
+            .is_ok();
             self.condition_cache.insert(condition.to_owned(), test_res);
             test_res
         }
@@ -1002,7 +995,7 @@ impl<'ctx> Completer<'ctx> {
         // search if we know the location of the whatis database. This can take some time on slower
         // systems with a large set of manuals, but it should be ok since apropos is only called once.
         let mut list = vec![];
-        exec_subshell(
+        let _ = exec_subshell(
             &lookup_cmd,
             parser,
             Some(&mut list),
@@ -1176,14 +1169,12 @@ impl<'ctx> Completer<'ctx> {
     fn complete_from_args(&mut self, s: &wstr, args: &wstr, desc: &wstr, flags: CompleteFlags) {
         let is_autosuggest = self.flags.autosuggestion;
 
-        let saved_state = if let Some(parser) = self.ctx.maybe_parser() {
-            let saved_interactive = parser.libdata().is_interactive;
-            parser.libdata_mut().is_interactive = false;
-
-            Some((saved_interactive, parser.get_last_statuses()))
-        } else {
-            None
-        };
+        let mut saved_statuses = None;
+        let mut scope = None;
+        if let Some(parser) = self.ctx.maybe_parser() {
+            saved_statuses = Some(parser.get_last_statuses());
+            scope = Some(parser.push_scope(|s| s.is_interactive = false));
+        }
 
         let eflags = if is_autosuggest {
             ExpandFlags::FAIL_ON_CMDSUBST
@@ -1194,10 +1185,9 @@ impl<'ctx> Completer<'ctx> {
         let possible_comp = Parser::expand_argument_list(args, eflags, self.ctx);
 
         if let Some(parser) = self.ctx.maybe_parser() {
-            let (saved_interactive, status) = saved_state.unwrap();
-            parser.libdata_mut().is_interactive = saved_interactive;
-            parser.set_last_statuses(status);
+            parser.set_last_statuses(saved_statuses.unwrap());
         }
+        std::mem::drop(scope);
 
         // Allow leading dots - see #3707.
         self.complete_strings(
@@ -1320,7 +1310,7 @@ impl<'ctx> Completer<'ctx> {
                                 }
                                 let (arg_prefix, arg) = s.split_once(arg_offset);
                                 let first_new = self.completions.completions.len();
-                                self.complete_from_args(arg, &o.comp, o.localized_desc(), o.flags);
+                                self.complete_from_args(arg, &o.comp, o.desc.localize(), o.flags);
                                 for compl in &mut self.completions.completions[first_new..] {
                                     if compl.replaces_token() {
                                         compl.completion.insert_utfstr(0, arg_prefix);
@@ -1351,7 +1341,7 @@ impl<'ctx> Completer<'ctx> {
                             if o.result_mode.force_files {
                                 has_force = true;
                             }
-                            self.complete_from_args(s, &o.comp, o.localized_desc(), o.flags);
+                            self.complete_from_args(s, &o.comp, o.desc.localize(), o.flags);
                         }
                     }
 
@@ -1387,7 +1377,7 @@ impl<'ctx> Completer<'ctx> {
                                 if o.result_mode.force_files {
                                     has_force = true;
                                 }
-                                self.complete_from_args(s, &o.comp, o.localized_desc(), o.flags);
+                                self.complete_from_args(s, &o.comp, o.desc.localize(), o.flags);
                             }
                         }
                     }
@@ -1410,7 +1400,7 @@ impl<'ctx> Completer<'ctx> {
                 if o.option.is_empty() {
                     use_files &= !o.result_mode.no_files;
                     has_force |= o.result_mode.force_files;
-                    self.complete_from_args(s, &o.comp, o.localized_desc(), o.flags);
+                    self.complete_from_args(s, &o.comp, o.desc.localize(), o.flags);
                 }
 
                 if !use_switches || s.is_empty() {
@@ -1443,7 +1433,7 @@ impl<'ctx> Completer<'ctx> {
                         }
                     }
                     // It's a match.
-                    let desc = o.localized_desc();
+                    let desc = o.desc.localize();
                     // Append a short-style option
                     if !self
                         .completions
@@ -1496,7 +1486,7 @@ impl<'ctx> Completer<'ctx> {
                     // Append a long-style option with a mandatory trailing equal sign
                     if !self.completions.add(Completion::new(
                         completion,
-                        o.localized_desc().to_owned(),
+                        o.desc.localize().to_owned(),
                         StringFuzzyMatch::exact_match(),
                         flags | CompleteFlags::NO_SPACE,
                     )) {
@@ -1507,7 +1497,7 @@ impl<'ctx> Completer<'ctx> {
                 // Append a long-style option
                 if !self.completions.add(Completion::new(
                     whole_opt.slice_from(offset).to_owned(),
-                    o.localized_desc().to_owned(),
+                    o.desc.localize().to_owned(),
                     StringFuzzyMatch::exact_match(),
                     flags,
                 )) {
@@ -1663,16 +1653,13 @@ impl<'ctx> Completer<'ctx> {
                 continue;
             };
 
-            let (comp, flags) = if !r#match.requires_full_replacement() {
+            let mut flags = CompleteFlags::VARIABLE_NAME;
+            let comp = if !r#match.requires_full_replacement() {
                 // Take only the suffix.
-                (
-                    env_name.slice_from(varlen).to_owned(),
-                    CompleteFlags::empty(),
-                )
+                env_name.slice_from(varlen).to_owned()
             } else {
-                let comp = whole_var.slice_to(start_offset).to_owned() + env_name.as_utfstr();
-                let flags = CompleteFlags::REPLACES_TOKEN | CompleteFlags::DONT_ESCAPE;
-                (comp, flags)
+                flags |= CompleteFlags::REPLACES_TOKEN | CompleteFlags::DONT_ESCAPE;
+                whole_var.slice_to(start_offset).to_owned() + env_name.as_utfstr()
             };
 
             let mut desc = WString::new();
@@ -1787,7 +1774,7 @@ impl<'ctx> Completer<'ctx> {
             // The getpwent() function does not exist on Android. A Linux user on Android isn't
             // really a user - each installed app gets an UID assigned. Listing all UID:s is not
             // possible without root access, and doing a ~USER type expansion does not make sense
-            // since every app is sandboxed and can't access eachother.
+            // since every app is sandboxed and can't access each other.
             return false;
         }
         #[cfg(not(target_os = "android"))]
@@ -1868,7 +1855,7 @@ impl<'ctx> Completer<'ctx> {
     fn apply_var_assignments<T: AsRef<wstr>>(
         &mut self,
         var_assignments: &[T],
-    ) -> Option<ScopeGuard<(), impl FnOnce(&mut ()) + 'ctx>> {
+    ) -> Option<ScopeGuard<(), impl FnOnce(()) + 'ctx>> {
         if !self.ctx.has_parser() || var_assignments.is_empty() {
             return None;
         }
@@ -1937,12 +1924,12 @@ impl<'ctx> Completer<'ctx> {
         // we're doing autosuggestions.
         let _remove_transient = (!is_autosuggest).then(|| {
             let parser = self.ctx.parser();
-            parser
+            let saved_transient = parser
                 .libdata_mut()
-                .transient_commandlines
-                .push(cmdline.to_owned());
+                .transient_commandline
+                .replace(cmdline.to_owned());
             ScopeGuard::new((), move |_| {
-                parser.libdata_mut().transient_commandlines.pop();
+                parser.libdata_mut().transient_commandline = saved_transient;
             })
         });
 
@@ -2166,7 +2153,7 @@ fn parse_cmd_string(s: &wstr, vars: &dyn Environment) -> CmdString {
 /// Returns a description for the specified function, or an empty string if none.
 fn complete_function_desc(f: &wstr) -> WString {
     if let Some(props) = function::get_props(f) {
-        props.description.clone()
+        props.description.localize().to_owned()
     } else {
         WString::new()
     }
@@ -2316,7 +2303,9 @@ pub fn complete_add(
         typ: option_type,
         result_mode,
         comp,
-        desc,
+        // The external source is a completion script in `share`,
+        // from which `build_tools/fish_xgettext.fish` extracts descriptions.
+        desc: LocalizableString::from_external_source(desc),
         conditions: condition,
         flags,
     };
@@ -2425,7 +2414,7 @@ fn completion2string(index: &CompletionEntryIndex, o: &CompleteEntryOpt) -> WStr
         CompleteOptionType::DoubleLong => append_switch_short_arg(&mut out, 'l', &o.option),
     }
 
-    append_switch_short_arg(&mut out, 'd', o.localized_desc());
+    append_switch_short_arg(&mut out, 'd', o.desc.localize());
     append_switch_short_arg(&mut out, 'a', &o.comp);
     for c in &o.conditions {
         append_switch_short_arg(&mut out, 'n', c);
