@@ -2,12 +2,13 @@ use super::environment_impl::{
     colon_split, uvars, EnvMutex, EnvMutexGuard, EnvScopedImpl, EnvStackImpl, ModResult,
     UVAR_SCOPE_IS_GLOBAL,
 };
-use super::{ConfigPaths, ElectricVar};
+use super::ElectricVar;
 use crate::abbrs::{abbrs_get_set, Abbreviation, Position};
+use crate::builtins::shared::{BuiltinResult, SUCCESS};
 use crate::common::{str2wcstring, unescape_string, wcs2zstring, UnescapeStringStyle};
+use crate::env::config_paths::ConfigPaths;
 use crate::env::{EnvMode, EnvVar, Statuses};
 use crate::env_dispatch::{env_dispatch_init, env_dispatch_var_change};
-use crate::env_universal_common::CallbackDataList;
 use crate::event::Event;
 use crate::flog::FLOG;
 use crate::global_safety::RelaxedAtomicBool;
@@ -27,8 +28,8 @@ use crate::wcstringutil::join_strings;
 use crate::wutil::{fish_wcstol, wgetcwd, wgettext};
 use std::sync::atomic::Ordering;
 
-use libc::{c_int, uid_t, STDOUT_FILENO, _IONBF};
-use once_cell::sync::OnceCell;
+use libc::{c_int, confstr, uid_t, STDOUT_FILENO, _IONBF};
+use once_cell::sync::{Lazy, OnceCell};
 use std::collections::HashMap;
 use std::ffi::CStr;
 use std::io::Write;
@@ -55,14 +56,14 @@ impl Default for EnvStackSetResult {
     }
 }
 
-impl From<EnvStackSetResult> for c_int {
+impl From<EnvStackSetResult> for BuiltinResult {
     fn from(r: EnvStackSetResult) -> Self {
         match r {
-            EnvStackSetResult::Ok => 0,
-            EnvStackSetResult::Perm => 1,
-            EnvStackSetResult::Scope => 2,
-            EnvStackSetResult::Invalid => 3,
-            EnvStackSetResult::NotFound => 4,
+            EnvStackSetResult::Ok => Ok(SUCCESS),
+            EnvStackSetResult::Perm => Err(1),
+            EnvStackSetResult::Scope => Err(2),
+            EnvStackSetResult::Invalid => Err(3),
+            EnvStackSetResult::NotFound => Err(4),
         }
     }
 }
@@ -164,7 +165,7 @@ impl EnvScoped {
         EnvScoped { inner }
     }
 
-    fn lock(&self) -> EnvMutexGuard<EnvScopedImpl> {
+    fn lock(&self) -> EnvMutexGuard<'_, EnvScopedImpl> {
         self.inner.lock()
     }
 }
@@ -199,7 +200,7 @@ impl EnvStack {
         }
     }
 
-    fn lock(&self) -> EnvMutexGuard<EnvStackImpl> {
+    fn lock(&self) -> EnvMutexGuard<'_, EnvStackImpl> {
         self.inner.lock()
     }
 
@@ -317,7 +318,7 @@ impl EnvStack {
         assert!(self.can_push_pop, "push/pop not allowed on global stack");
         let popped = self.lock().pop();
         if self.dispatches_var_changes {
-            // TODO: we would like to coalesce locale / curses changes, so that we only re-initialize
+            // TODO: we would like to coalesce locale changes, so that we only re-initialize
             // once.
             for key in popped {
                 env_dispatch_var_change(&key, self);
@@ -350,22 +351,23 @@ impl EnvStack {
         }
         UVARS_LOCALLY_MODIFIED.store(false);
 
-        let mut callbacks = CallbackDataList::new();
-        let changed = uvars().sync(&mut callbacks);
+        let (changed, callbacks) = uvars().sync();
         if changed {
             default_notifier().post_notification();
         }
         // React internally to changes to special variables like LANG, and populate on-variable events.
         let mut result = Vec::new();
-        for callback in callbacks {
-            let name = callback.key;
-            env_dispatch_var_change(&name, self);
-            let evt = if callback.val.is_none() {
-                Event::variable_erase(name)
-            } else {
-                Event::variable_set(name)
-            };
-            result.push(evt);
+        if let Some(callbacks) = callbacks {
+            for callback in callbacks {
+                let name = callback.key;
+                env_dispatch_var_change(&name, self);
+                let evt = if callback.val.is_none() {
+                    Event::variable_erase(name)
+                } else {
+                    Event::variable_set(name)
+                };
+                result.push(evt);
+            }
         }
         result
     }
@@ -402,7 +404,7 @@ impl Environment for EnvScoped {
     }
 }
 
-/// Necessary for Arc<EnvStack> to be sync.
+/// Necessary for [`Arc<EnvStack>`] to be sync.
 /// Safety: again, the global lock.
 unsafe impl Send for EnvStack {}
 
@@ -566,29 +568,39 @@ fn setup_user(vars: &EnvStack) {
     }
 }
 
-/// Make sure the PATH variable contains something.
-fn setup_path() {
-    use crate::libc::{confstr, _CS_PATH};
-
-    let vars = EnvStack::globals();
-    let path = vars.get_unless_empty(L!("PATH"));
-    if path.is_none() {
-        // _CS_PATH: colon-separated paths to find POSIX utilities
-
-        let buf_size = unsafe { confstr(_CS_PATH(), std::ptr::null_mut(), 0) };
-        let path = if buf_size > 0 {
+pub(crate) static FALLBACK_PATH: Lazy<&[WString]> = Lazy::new(|| {
+    use crate::libc::_CS_PATH;
+    // _CS_PATH: colon-separated paths to find POSIX utilities
+    let buf_size = unsafe { confstr(_CS_PATH(), std::ptr::null_mut(), 0) };
+    Box::leak(
+        (if buf_size > 0 {
             let mut buf = vec![b'\0' as libc::c_char; buf_size];
             unsafe { confstr(_CS_PATH(), buf.as_mut_ptr(), buf_size) };
             let buf = buf;
             // safety: buf should contain a null-byte, and is not mutable unless we move ownership
             let cstr = unsafe { CStr::from_ptr(buf.as_ptr()) };
-            str2wcstring(cstr.to_bytes())
+            colon_split(&[str2wcstring(cstr.to_bytes())])
         } else {
-            // the above should really not fail
-            join_strings(crate::path::DEFAULT_PATH.as_ref(), ':')
-        };
+            vec![
+                WString::from_str(env!("PREFIX")) + L!("/bin"),
+                L!("/usr/bin").to_owned(),
+                L!("/bin").to_owned(),
+            ]
+        })
+        .into_boxed_slice(),
+    )
+});
 
-        vars.set_one(L!("PATH"), EnvMode::GLOBAL | EnvMode::EXPORT, path);
+/// Make sure the PATH variable contains something.
+fn setup_path() {
+    let vars = EnvStack::globals();
+    let path = vars.get_unless_empty(L!("PATH"));
+    if path.is_none() {
+        vars.set(
+            L!("PATH"),
+            EnvMode::GLOBAL | EnvMode::EXPORT,
+            FALLBACK_PATH.to_vec(),
+        );
     }
 }
 
@@ -634,24 +646,36 @@ pub fn env_init(paths: Option<&ConfigPaths>, do_uvars: bool, default_paths: bool
     setup_user(vars);
 
     if let Some(paths) = paths {
-        let datadir = str2wcstring(paths.data.as_os_str().as_bytes());
+        #[cfg(feature = "embed-data")]
+        vars.set_empty(FISH_DATADIR_VAR, EnvMode::GLOBAL);
+        #[cfg(not(feature = "embed-data"))]
+        {
+            let datadir = str2wcstring(paths.data.as_os_str().as_bytes());
+            vars.set_one(FISH_DATADIR_VAR, EnvMode::GLOBAL, datadir.clone());
+
+            if default_paths {
+                let mut scstr = datadir;
+                // This is generated by PathBuf.join() everywhere currently
+                assert!(!scstr.ends_with("/"));
+                scstr.push_str("/functions");
+                vars.set_one(L!("fish_function_path"), EnvMode::GLOBAL, scstr);
+            }
+        }
 
-        vars.set_one(FISH_DATADIR_VAR, EnvMode::GLOBAL, datadir.clone());
         vars.set_one(
             FISH_SYSCONFDIR_VAR,
             EnvMode::GLOBAL,
             str2wcstring(paths.sysconf.as_os_str().as_bytes()),
         );
 
-        if !cfg!(feature = "installable") {
-            vars.set_one(
-                FISH_HELPDIR_VAR,
-                EnvMode::GLOBAL,
-                str2wcstring(paths.doc.as_os_str().as_bytes()),
-            );
-        } else {
-            vars.set_empty(FISH_HELPDIR_VAR, EnvMode::GLOBAL);
-        }
+        #[cfg(feature = "embed-data")]
+        vars.set_empty(FISH_HELPDIR_VAR, EnvMode::GLOBAL);
+        #[cfg(not(feature = "embed-data"))]
+        vars.set_one(
+            FISH_HELPDIR_VAR,
+            EnvMode::GLOBAL,
+            str2wcstring(paths.doc.as_os_str().as_bytes()),
+        );
         if let Some(bp) = &paths.bin {
             vars.set_one(
                 FISH_BIN_DIR,
@@ -661,14 +685,6 @@ pub fn env_init(paths: Option<&ConfigPaths>, do_uvars: bool, default_paths: bool
         } else {
             vars.set_empty(FISH_BIN_DIR, EnvMode::GLOBAL);
         };
-
-        if default_paths {
-            let mut scstr = datadir;
-            // This is generated by PathBuf.join() everywhere currently
-            assert!(!scstr.ends_with("/"));
-            scstr.push_str("/functions");
-            vars.set_one(L!("fish_function_path"), EnvMode::GLOBAL, scstr);
-        }
     }
 
     let user_config_dir = path_get_config();
@@ -760,7 +776,6 @@ pub fn env_init(paths: Option<&ConfigPaths>, do_uvars: bool, default_paths: bool
     }
 
     // Initialize termsize variables.
-    // PORTING: 3x deref is weird
     let termsize = termsize::SHARED_CONTAINER.initialize(vars as &dyn Environment);
     if vars.get_unless_empty(L!("COLUMNS")).is_none() {
         vars.set_one(L!("COLUMNS"), EnvMode::GLOBAL, termsize.width.to_wstring());
@@ -787,8 +802,7 @@ pub fn env_init(paths: Option<&ConfigPaths>, do_uvars: bool, default_paths: bool
         UVAR_SCOPE_IS_GLOBAL.store(true);
     } else {
         // Set up universal variables using the default path.
-        let mut callbacks = CallbackDataList::new();
-        uvars().initialize(&mut callbacks);
+        let callbacks = uvars().initialize().unwrap_or_default();
         for callback in callbacks {
             env_dispatch_var_change(&callback.key, vars);
         }
@@ -805,7 +819,7 @@ pub fn env_init(paths: Option<&ConfigPaths>, do_uvars: bool, default_paths: bool
 
                 // Look for a global exported variable with the same name.
                 let global = EnvStack::globals().getf(name, EnvMode::GLOBAL | EnvMode::EXPORT);
-                if global.is_some() && global.unwrap().as_string() == uvar.as_string() {
+                if global.is_some_and(|x| x.as_string() == uvar.as_string()) {
                     to_skip.push(name.to_owned());
                 }
             }
