@@ -4,7 +4,6 @@ use crate::wcstringutil::join_strings;
 use bitflags::bitflags;
 use libc::c_int;
 use std::collections::HashMap;
-use std::path::PathBuf;
 use std::sync::Arc;
 
 /// The character used to delimit path and non-path variables in exporting and in string expansion.
@@ -46,17 +45,6 @@ impl From<EnvMode> for u16 {
     }
 }
 
-/// A struct of configuration directories, determined in main() that fish will optionally pass to
-/// env_init.
-#[derive(Default)]
-pub struct ConfigPaths {
-    pub data: PathBuf,           // e.g., /usr/local/share
-    pub sysconf: PathBuf,        // e.g., /usr/local/etc
-    pub doc: PathBuf,            // e.g., /usr/local/share/doc/fish
-    pub bin: Option<PathBuf>,    // e.g., /usr/local/bin
-    pub locale: Option<PathBuf>, // e.g., /usr/local/share/locale
-}
-
 /// A collection of status and pipestatus.
 #[derive(Clone, Debug)]
 pub struct Statuses {
@@ -103,7 +91,7 @@ bitflags! {
 pub struct EnvVar {
     /// The list of values in this variable.
     /// Arc allows for cheap copying
-    values: Arc<Box<[WString]>>,
+    values: Arc<[WString]>,
     /// The variable's flags.
     flags: EnvVarFlags,
 }
@@ -112,8 +100,8 @@ impl Default for EnvVar {
     fn default() -> Self {
         use std::sync::OnceLock;
         /// A shared read-only empty list.
-        static EMPTY_LIST: OnceLock<Arc<Box<[WString]>>> = OnceLock::new();
-        let empty_list = EMPTY_LIST.get_or_init(|| Arc::new(Box::new([])));
+        static EMPTY_LIST: OnceLock<Arc<[WString]>> = OnceLock::new();
+        let empty_list = EMPTY_LIST.get_or_init(|| Arc::new([]));
 
         EnvVar {
             values: Arc::clone(empty_list),
@@ -131,7 +119,7 @@ impl EnvVar {
     /// Creates a new `EnvVar`.
     pub fn new_vec(values: Vec<WString>, flags: EnvVarFlags) -> Self {
         EnvVar {
-            values: Arc::new(values.into_boxed_slice()),
+            values: values.into(),
             flags,
         }
     }
@@ -200,15 +188,15 @@ impl EnvVar {
     }
 
     /// Returns a copy of the variable with new values.
-    pub fn setting_vals(&mut self, values: Vec<WString>) -> Self {
+    pub fn setting_vals(&self, values: Vec<WString>) -> Self {
         EnvVar {
-            values: Arc::new(values.into_boxed_slice()),
+            values: values.into(),
             flags: self.flags,
         }
     }
 
     /// Returns a copy of the variable with the export flag changed.
-    pub fn setting_exports(&mut self, export: bool) -> Self {
+    pub fn setting_exports(&self, export: bool) -> Self {
         let mut flags = self.flags;
         flags.set(EnvVarFlags::EXPORT, export);
         EnvVar {
@@ -218,7 +206,7 @@ impl EnvVar {
     }
 
     /// Returns a copy of the variable with the path variable flag changed.
-    pub fn setting_pathvar(&mut self, pathvar: bool) -> Self {
+    pub fn setting_pathvar(&self, pathvar: bool) -> Self {
         let mut flags = self.flags;
         flags.set(EnvVarFlags::PATHVAR, pathvar);
         EnvVar {
