@@ -7,7 +7,7 @@
 use std::sync::atomic::{AtomicBool, AtomicU32, Ordering};
 use std::sync::{Arc, Mutex};
 
-use crate::common::{escape, scoped_push_replacer, ScopeGuard};
+use crate::common::{escape, ScopeGuard};
 use crate::flog::FLOG;
 use crate::io::{IoChain, IoStreams};
 use crate::job_group::MaybeJobId;
@@ -264,7 +264,7 @@ impl Event {
 
     /// Test if specified event is blocked.
     fn is_blocked(&self, parser: &Parser) -> bool {
-        for block in parser.blocks().iter().rev() {
+        for block in parser.blocks_iter_rev() {
             if block.event_blocks {
                 return true;
             }
@@ -371,7 +371,7 @@ pub fn is_signal_observed(sig: libc::c_int) -> bool {
     // We are in a signal handler!
     OBSERVED_SIGNALS
         .get(usize::try_from(sig).unwrap())
-        .map_or(false, |s| s.load(Ordering::Relaxed) > 0)
+        .is_some_and(|s| s.load(Ordering::Relaxed) > 0)
 }
 
 pub fn get_desc(parser: &Parser, evt: &Event) -> WString {
@@ -460,21 +460,11 @@ pub fn get_function_handlers(name: &wstr) -> EventHandlerList {
 /// event handler, we make sure to optimize the 'no matches' path. This means that nothing is
 /// allocated/initialized unless needed.
 fn fire_internal(parser: &Parser, event: &Event) {
-    assert!(
-        parser.libdata().is_event >= 0,
-        "is_event should not be negative"
-    );
-
     // Suppress fish_trace during events.
-    let is_event = parser.libdata().is_event;
-    let _inc_event = scoped_push_replacer(
-        |new_value| std::mem::replace(&mut parser.libdata_mut().is_event, new_value),
-        is_event + 1,
-    );
-    let _suppress_trace = scoped_push_replacer(
-        |new_value| std::mem::replace(&mut parser.libdata_mut().suppress_fish_trace, new_value),
-        true,
-    );
+    let _saved = parser.push_scope(|s| {
+        s.is_event = true;
+        s.suppress_fish_trace = true;
+    });
 
     // Capture the event handlers that match this event.
     let fire: Vec<_> = EVENT_HANDLERS
@@ -503,12 +493,10 @@ fn fire_internal(parser: &Parser, event: &Event) {
 
         // Event handlers are not part of the main flow of code, so they are marked as
         // non-interactive.
-        let saved_is_interactive =
-            std::mem::replace(&mut parser.libdata_mut().is_interactive, false);
+        let _non_interactive = parser.push_scope(|s| s.is_interactive = false);
         let saved_statuses = parser.get_last_statuses();
         let _cleanup = ScopeGuard::new((), |()| {
             parser.set_last_statuses(saved_statuses);
-            parser.libdata_mut().is_interactive = saved_is_interactive;
         });
 
         FLOG!(
@@ -536,10 +524,8 @@ fn fire_internal(parser: &Parser, event: &Event) {
 /// Fire all delayed events attached to the given parser.
 pub fn fire_delayed(parser: &Parser) {
     {
-        let ld = &parser.libdata();
-
         // Do not invoke new event handlers from within event handlers.
-        if ld.is_event != 0 {
+        if parser.scope().is_event {
             return;
         };
     }
