@@ -1,12 +1,13 @@
-use crate::common::{escape, get_by_sorted_name, shell_modes, str2wcstring, Named};
-use crate::curses;
-use crate::env::{Environment, CURSES_INITIALIZED};
+use crate::common::{escape, get_by_sorted_name, str2wcstring, Named};
+use crate::env::Environment;
 use crate::event;
 use crate::flog::FLOG;
+// Polyfill for Option::is_none_or(), stabilized in 1.82.0
 #[allow(unused_imports)]
 use crate::future::IsSomeAnd;
+use crate::global_safety::RelaxedAtomicBool;
 use crate::input_common::{
-    match_key_event_to_key, CharEvent, CharInputStyle, InputData, InputEventQueuer, KeyEvent,
+    match_key_event_to_key, CharEvent, CharInputStyle, ImplicitEvent, InputData, InputEventQueuer,
     KeyMatchQuality, ReadlineCmd, R_END_INPUT_FUNCTIONS,
 };
 use crate::key::{self, canonicalize_raw_escapes, ctrl, Key, Modifiers};
@@ -17,8 +18,7 @@ use crate::reader::{
 use crate::signal::signal_clear_cancel;
 use crate::threads::{assert_is_main_thread, iothread_service_main};
 use crate::wchar::prelude::*;
-use once_cell::sync::{Lazy, OnceCell};
-use std::ffi::CString;
+use once_cell::sync::Lazy;
 use std::mem;
 use std::sync::{
     atomic::{AtomicU32, Ordering},
@@ -41,7 +41,6 @@ pub struct InputMappingName {
 pub enum KeyNameStyle {
     Plain,
     RawEscapeSequence,
-    Terminfo(WString),
 }
 
 /// Struct representing a keybinding. Returned by input_get_mappings.
@@ -57,7 +56,7 @@ pub struct InputMapping {
     mode: WString,
     /// New mode that should be switched to after command evaluation, or None to leave the mode unchanged.
     sets_mode: Option<WString>,
-    /// Perhaps this binding was created using a raw escape sequence or terminfo.
+    /// Perhaps this binding was created using a raw escape sequence.
     key_name_style: KeyNameStyle,
 }
 
@@ -92,16 +91,6 @@ impl InputMapping {
     }
 }
 
-/// A struct representing the mapping from a terminfo key name to a terminfo character sequence.
-#[derive(Debug)]
-struct TerminfoMapping {
-    // name of key
-    name: &'static wstr,
-
-    // character sequence generated on keypress, or none if there was no mapping.
-    seq: Option<Box<[u8]>>,
-}
-
 /// Input function metadata. This list should be kept in sync with the key code list in
 /// input_common.rs.
 struct InputFunctionMetadata {
@@ -124,10 +113,6 @@ const fn make_md(name: &'static wstr, code: ReadlineCmd) -> InputFunctionMetadat
 /// Keep this list sorted alphabetically!
 #[rustfmt::skip]
 const INPUT_FUNCTION_METADATA: &[InputFunctionMetadata] = &[
-    // NULL makes it unusable - this is specially inserted when we detect mouse input
-    make_md(L!(""), ReadlineCmd::DisableMouseTracking),
-    make_md(L!(""), ReadlineCmd::FocusIn),
-    make_md(L!(""), ReadlineCmd::FocusOut),
     make_md(L!("accept-autosuggestion"), ReadlineCmd::AcceptAutosuggestion),
     make_md(L!("and"), ReadlineCmd::FuncAnd),
     make_md(L!("backward-bigword"), ReadlineCmd::BackwardBigword),
@@ -158,6 +143,7 @@ const INPUT_FUNCTION_METADATA: &[InputFunctionMetadata] = &[
     make_md(L!("delete-char"), ReadlineCmd::DeleteChar),
     make_md(L!("delete-or-exit"), ReadlineCmd::DeleteOrExit),
     make_md(L!("down-line"), ReadlineCmd::DownLine),
+    make_md(L!("downcase-selection"), ReadlineCmd::DowncaseSelection),
     make_md(L!("downcase-word"), ReadlineCmd::DowncaseWord),
     make_md(L!("end-of-buffer"), ReadlineCmd::EndOfBuffer),
     make_md(L!("end-of-history"), ReadlineCmd::EndOfHistory),
@@ -176,7 +162,11 @@ const INPUT_FUNCTION_METADATA: &[InputFunctionMetadata] = &[
     make_md(L!("forward-single-char"), ReadlineCmd::ForwardSingleChar),
     make_md(L!("forward-token"), ReadlineCmd::ForwardToken),
     make_md(L!("forward-word"), ReadlineCmd::ForwardWord),
+    make_md(L!("history-delete"), ReadlineCmd::HistoryDelete),
+    make_md(L!("history-last-token-search-backward"), ReadlineCmd::HistoryLastTokenSearchBackward),
+    make_md(L!("history-last-token-search-forward"), ReadlineCmd::HistoryLastTokenSearchForward),
     make_md(L!("history-pager"), ReadlineCmd::HistoryPager),
+    #[allow(deprecated)]
     make_md(L!("history-pager-delete"), ReadlineCmd::HistoryPagerDelete),
     make_md(L!("history-prefix-search-backward"), ReadlineCmd::HistoryPrefixSearchBackward),
     make_md(L!("history-prefix-search-forward"), ReadlineCmd::HistoryPrefixSearchForward),
@@ -204,6 +194,7 @@ const INPUT_FUNCTION_METADATA: &[InputFunctionMetadata] = &[
     make_md(L!("repaint-mode"), ReadlineCmd::RepaintMode),
     make_md(L!("repeat-jump"), ReadlineCmd::RepeatJump),
     make_md(L!("repeat-jump-reverse"), ReadlineCmd::ReverseRepeatJump),
+    make_md(L!("scrollback-push"), ReadlineCmd::ScrollbackPush),
     make_md(L!("self-insert"), ReadlineCmd::SelfInsert),
     make_md(L!("self-insert-notfirst"), ReadlineCmd::SelfInsertNotFirst),
     make_md(L!("suppress-autosuggestion"), ReadlineCmd::SuppressAutosuggestion),
@@ -214,6 +205,7 @@ const INPUT_FUNCTION_METADATA: &[InputFunctionMetadata] = &[
     make_md(L!("transpose-words"), ReadlineCmd::TransposeWords),
     make_md(L!("undo"), ReadlineCmd::Undo),
     make_md(L!("up-line"), ReadlineCmd::UpLine),
+    make_md(L!("upcase-selection"), ReadlineCmd::UpcaseSelection),
     make_md(L!("upcase-word"), ReadlineCmd::UpcaseWord),
     make_md(L!("yank"), ReadlineCmd::Yank),
     make_md(L!("yank-pop"), ReadlineCmd::YankPop),
@@ -256,9 +248,6 @@ pub fn input_mappings() -> MutexGuard<'static, InputMappingSet> {
     INPUT_MAPPINGS.lock().unwrap()
 }
 
-/// Terminfo map list.
-static TERMINFO_MAPPINGS: OnceCell<Box<[TerminfoMapping]>> = OnceCell::new();
-
 /// Return the current bind mode.
 fn input_get_bind_mode(vars: &dyn Environment) -> WString {
     if let Some(mode) = vars.get(FISH_BIND_MODE_VAR) {
@@ -345,10 +334,11 @@ impl InputMappingSet {
 /// initializations for our input subsystem.
 pub fn init_input() {
     assert_is_main_thread();
-    if TERMINFO_MAPPINGS.get().is_some() {
+
+    static DONE: RelaxedAtomicBool = RelaxedAtomicBool::new(false);
+    if DONE.swap(true) {
         return;
     }
-    TERMINFO_MAPPINGS.set(create_input_terminfo()).unwrap();
 
     let mut input_mapping = input_mappings();
 
@@ -431,10 +421,7 @@ impl<'a> InputEventQueuer for Reader<'a> {
 
         // Tell the reader an event occurred.
         if reader_reading_interrupted(self) != 0 {
-            let vintr = shell_modes().c_cc[libc::VINTR];
-            if vintr != 0 {
-                self.push_front(CharEvent::from_key(KeyEvent::from_single_byte(vintr)));
-            }
+            self.enqueue_interrupt_key();
             return;
         }
         self.push_front(CharEvent::from_check_exit());
@@ -628,9 +615,10 @@ impl<'q, Queuer: InputEventQueuer + ?Sized> EventQueuePeeker<'q, Queuer> {
 
     /// Test if any of our peeked events are readline or check_exit.
     fn char_sequence_interrupted(&self) -> bool {
-        self.peeked
-            .iter()
-            .any(|evt| evt.is_readline_or_command() || evt.is_check_exit())
+        self.peeked.iter().any(|evt| {
+            evt.is_readline_or_command()
+                || matches!(evt, CharEvent::Implicit(ImplicitEvent::CheckExit))
+        })
     }
 
     /// Reset our index back to 0.
@@ -782,7 +770,15 @@ impl<'a> Reader<'a> {
                         self.insert_front(seq);
                         // Issue #1595: ensure we only insert characters, not readline functions. The
                         // common case is that this will be empty.
-                        let mut res = self.read_characters_no_readline();
+                        let mut res = self.read_character_matching(|evt| {
+                            use CharEvent::*;
+                            use ImplicitEvent::*;
+                            match evt {
+                                Key(_) => true,
+                                Implicit(Eof) => true,
+                                Readline(_) | Command(_) | Implicit(_) | QueryResult(_) => false,
+                            }
+                        });
 
                         // Hackish: mark the input style.
                         if readline_event.cmd == ReadlineCmd::SelfInsertNotFirst {
@@ -808,15 +804,6 @@ impl<'a> Reader<'a> {
                 CharEvent::Command(_) => {
                     return evt;
                 }
-                CharEvent::Eof => {
-                    // If we have EOF, we need to immediately quit.
-                    // There's no need to go through the input functions.
-                    return evt;
-                }
-                CharEvent::CheckExit => {
-                    // Allow the reader to check for exit conditions.
-                    return evt;
-                }
                 CharEvent::Key(ref kevt) => {
                     FLOG!(
                         reader,
@@ -831,16 +818,19 @@ impl<'a> Reader<'a> {
                     self.push_front(evt);
                     self.mapping_execute_matching_or_generic();
                 }
+                CharEvent::Implicit(_) | CharEvent::QueryResult(_) => {
+                    return evt;
+                }
             }
         }
     }
 
     fn mapping_execute_matching_or_generic(&mut self) {
-        let vars = self.parser.vars_ref();
+        let vars = self.parser.vars();
         let mut peeker = EventQueuePeeker::new(self);
         // Check for ordinary mappings.
         let ip = input_mappings();
-        if let Some(mapping) = peeker.find_mapping(&*vars, &ip) {
+        if let Some(mapping) = peeker.find_mapping(vars, &ip) {
             FLOG!(
                 reader,
                 format!("Found mapping {:?} from {:?}", &mapping, &peeker.peeked)
@@ -866,22 +856,18 @@ impl<'a> Reader<'a> {
         peeker.consume();
     }
 
-    /// Helper function. Picks through the queue of incoming characters until we get to one that's not a
-    /// readline function.
-    fn read_characters_no_readline(&mut self) -> CharEvent {
+    /// Pick through the queue of incoming characters until we get to one that matches.
+    fn read_character_matching(&mut self, predicate: impl Fn(&CharEvent) -> bool) -> CharEvent {
         let mut saved_events = std::mem::take(&mut self.get_input_data_mut().event_storage);
         assert!(saved_events.is_empty(), "event_storage should be empty");
 
-        let evt_to_return: CharEvent;
-        loop {
+        let evt_to_return: CharEvent = loop {
             let evt = self.readch();
-            if evt.is_readline_or_command() || evt.is_check_exit() {
-                saved_events.push(evt);
-            } else {
-                evt_to_return = evt;
-                break;
+            if (predicate)(&evt) {
+                break evt;
             }
-        }
+            saved_events.push(evt);
+        };
 
         // Restore any readline functions
         self.insert_front(saved_events.drain(..));
@@ -992,7 +978,7 @@ impl InputMappingSet {
         } else {
             &mut self.preset_mapping_list
         };
-        let should_erase = |m: &InputMapping| mode.is_none() || mode.unwrap() == m.mode;
+        let should_erase = |m: &InputMapping| mode.is_none_or(|x| x == m.mode);
         ml.retain(|m| !should_erase(m));
     }
 
@@ -1042,125 +1028,6 @@ impl InputMappingSet {
     }
 }
 
-/// Create a list of terminfo mappings.
-fn create_input_terminfo() -> Box<[TerminfoMapping]> {
-    assert!(CURSES_INITIALIZED.load(Ordering::Relaxed));
-    let Some(term) = curses::term() else {
-        // setupterm() failed so we can't reference any key definitions.
-        return Box::new([]);
-    };
-
-    // Helper to convert an Option<CString> to an Option<Box<[u8]>>.
-    // The nul-terminator is NOT included.
-    fn opt_cstr_to_bytes(opt: &Option<CString>) -> Option<Box<[u8]>> {
-        opt.clone().map(|s| s.into_bytes().into())
-    }
-
-    macro_rules! terminfo_add {
-        ($key:ident) => {
-            TerminfoMapping {
-                name: &L!(stringify!($key))[4..],
-                seq: opt_cstr_to_bytes(&term.$key),
-            }
-        };
-    }
-    #[rustfmt::skip]
-    return Box::new([
-        terminfo_add!(key_a1), terminfo_add!(key_a3), terminfo_add!(key_b2),
-        terminfo_add!(key_backspace), terminfo_add!(key_beg), terminfo_add!(key_btab),
-        terminfo_add!(key_c1), terminfo_add!(key_c3), terminfo_add!(key_cancel),
-        terminfo_add!(key_catab), terminfo_add!(key_clear), terminfo_add!(key_close),
-        terminfo_add!(key_command), terminfo_add!(key_copy), terminfo_add!(key_create),
-        terminfo_add!(key_ctab), terminfo_add!(key_dc), terminfo_add!(key_dl), terminfo_add!(key_down),
-        terminfo_add!(key_eic), terminfo_add!(key_end), terminfo_add!(key_enter),
-        terminfo_add!(key_eol), terminfo_add!(key_eos), terminfo_add!(key_exit), terminfo_add!(key_f0),
-        terminfo_add!(key_f1), terminfo_add!(key_f2), terminfo_add!(key_f3), terminfo_add!(key_f4),
-        terminfo_add!(key_f5), terminfo_add!(key_f6), terminfo_add!(key_f7), terminfo_add!(key_f8),
-        terminfo_add!(key_f9), terminfo_add!(key_f10), terminfo_add!(key_f11), terminfo_add!(key_f12),
-        terminfo_add!(key_find), terminfo_add!(key_help), terminfo_add!(key_home),
-        terminfo_add!(key_ic), terminfo_add!(key_il), terminfo_add!(key_left), terminfo_add!(key_ll),
-        terminfo_add!(key_mark), terminfo_add!(key_message), terminfo_add!(key_move),
-        terminfo_add!(key_next), terminfo_add!(key_npage), terminfo_add!(key_open),
-        terminfo_add!(key_options), terminfo_add!(key_ppage), terminfo_add!(key_previous),
-        terminfo_add!(key_print), terminfo_add!(key_redo), terminfo_add!(key_reference),
-        terminfo_add!(key_refresh), terminfo_add!(key_replace), terminfo_add!(key_restart),
-        terminfo_add!(key_resume), terminfo_add!(key_right), terminfo_add!(key_save),
-        terminfo_add!(key_sbeg), terminfo_add!(key_scancel), terminfo_add!(key_scommand),
-        terminfo_add!(key_scopy), terminfo_add!(key_screate), terminfo_add!(key_sdc),
-        terminfo_add!(key_sdl), terminfo_add!(key_select), terminfo_add!(key_send),
-        terminfo_add!(key_seol), terminfo_add!(key_sexit), terminfo_add!(key_sf),
-        terminfo_add!(key_sfind), terminfo_add!(key_shelp), terminfo_add!(key_shome),
-        terminfo_add!(key_sic), terminfo_add!(key_sleft), terminfo_add!(key_smessage),
-        terminfo_add!(key_smove), terminfo_add!(key_snext), terminfo_add!(key_soptions),
-        terminfo_add!(key_sprevious), terminfo_add!(key_sprint), terminfo_add!(key_sr),
-        terminfo_add!(key_sredo), terminfo_add!(key_sreplace), terminfo_add!(key_sright),
-        terminfo_add!(key_srsume), terminfo_add!(key_ssave), terminfo_add!(key_ssuspend),
-        terminfo_add!(key_stab), terminfo_add!(key_sundo), terminfo_add!(key_suspend),
-        terminfo_add!(key_undo), terminfo_add!(key_up),
-
-        // We introduce our own name for the string containing only the nul character - see
-        // #3189. This can typically be generated via control-space.
-        TerminfoMapping { name: NUL_MAPPING_NAME, seq: Some(Box::new([0])) },
-    ]);
-}
-
-/// Possible errors from from input_terminfo_get_sequence.
-pub enum GetSequenceError {
-    /// The mapping was not found.
-    NotFound,
-    /// The terminfo variable does not have a value.
-    NoSeq,
-}
-
-/// Return the sequence for the terminfo variable of the specified name.
-///
-/// If no terminfo variable of the specified name could be found, return false and set errno to
-/// ENOENT. If the terminfo variable does not have a value, return false and set errno to EILSEQ.
-pub fn input_terminfo_get_sequence(name: &wstr) -> Result<WString, GetSequenceError> {
-    let mappings = TERMINFO_MAPPINGS
-        .get()
-        .expect("TERMINFO_MAPPINGS not initialized");
-    for m in mappings.iter() {
-        if name == m.name {
-            // Found the mapping.
-            return if let Some(seq) = &m.seq {
-                Ok(str2wcstring(seq))
-            } else {
-                Err(GetSequenceError::NoSeq)
-            };
-        }
-    }
-    Err(GetSequenceError::NotFound)
-}
-
-/// Return the name of the terminfo variable with the specified sequence.
-pub fn input_terminfo_get_name(seq: &wstr) -> Option<WString> {
-    let mappings = TERMINFO_MAPPINGS
-        .get()
-        .expect("TERMINFO_MAPPINGS not initialized");
-    for m in mappings.iter() {
-        if m.seq.is_some() && seq == str2wcstring(m.seq.as_ref().unwrap()) {
-            return Some(m.name.to_owned());
-        }
-    }
-    None
-}
-
-/// Return a list of all known terminfo names.
-pub fn input_terminfo_get_names(skip_null: bool) -> Vec<WString> {
-    let mappings = TERMINFO_MAPPINGS
-        .get()
-        .expect("TERMINFO_MAPPINGS not initialized");
-    let mut result = Vec::with_capacity(mappings.len());
-    for m in mappings.iter() {
-        if skip_null && m.seq.is_none() {
-            continue;
-        }
-        result.push(m.name.to_owned());
-    }
-    result
-}
-
 /// Returns a list of all existing input function names.
 pub fn input_function_get_names() -> Vec<&'static wstr> {
     // Note: the C++ cached this, but we don't to save memory.
