@@ -17,7 +17,7 @@ bitflags! {
         const CONTINUE_AFTER_ERROR = 1 << 0;
         /// include comment tokens.
         const INCLUDE_COMMENTS = 1 << 1;
-        /// indicate that the tokenizer should accept incomplete tokens */
+        /// indicate that the tokenizer should accept incomplete tokens
         const ACCEPT_INCOMPLETE_TOKENS = 1 << 2;
         /// indicate that the parser should not generate the terminate token, allowing an 'unfinished'
         /// tree where some nodes may have no productions.
@@ -38,15 +38,15 @@ bitflags! {
 }
 
 /// A range of source code.
-#[derive(PartialEq, Eq, Clone, Copy, Debug)]
+#[derive(PartialEq, Eq, Clone, Copy, Debug, Default)]
 pub struct SourceRange {
     pub start: u32,
     pub length: u32,
 }
 
 impl SourceRange {
-    pub fn as_usize(&self) -> std::ops::Range<usize> {
-        (*self).into()
+    pub fn as_usize(self) -> std::ops::Range<usize> {
+        self.into()
     }
 }
 
@@ -57,6 +57,8 @@ pub enum ParseTokenType {
     // Terminal types.
     string,
     pipe,
+    left_brace,
+    right_brace,
     redirection,
     background,
     andand,
@@ -70,30 +72,29 @@ pub enum ParseTokenType {
     comment,
 }
 
-#[repr(u8)]
 #[derive(Clone, Copy, Debug, Eq, PartialEq)]
 pub enum ParseKeyword {
-    // 'none' is not a keyword, it is a sentinel indicating nothing.
-    none,
-
-    kw_and,
-    kw_begin,
-    kw_builtin,
-    kw_case,
-    kw_command,
-    kw_else,
-    kw_end,
-    kw_exclam,
-    kw_exec,
-    kw_for,
-    kw_function,
-    kw_if,
-    kw_in,
-    kw_not,
-    kw_or,
-    kw_switch,
-    kw_time,
-    kw_while,
+    // 'None' is not a keyword, it is a sentinel indicating nothing.
+    // Note it proves convenient to keep this as a value rather than using Option.
+    None,
+    And,
+    Begin,
+    Builtin,
+    Case,
+    Command,
+    Else,
+    End,
+    Exclam,
+    Exec,
+    For,
+    Function,
+    If,
+    In,
+    Not,
+    Or,
+    Switch,
+    Time,
+    While,
 }
 
 // Statement decorations like 'command' or 'exec'.
@@ -126,6 +127,7 @@ pub enum ParseErrorCode {
     unbalancing_end,          // end outside of block
     unbalancing_else,         // else outside of if
     unbalancing_case,         // case outside of switch
+    unbalancing_brace,        // } outside of {
     bare_variable_assignment, // a=b without command
     andor_in_pipeline,        // "and" or "or" after a pipe
 }
@@ -145,20 +147,20 @@ impl SourceRange {
             length: length.try_into().unwrap(),
         }
     }
-    pub fn start(&self) -> usize {
+    pub fn start(self) -> usize {
         self.start.try_into().unwrap()
     }
-    pub fn length(&self) -> usize {
+    pub fn length(self) -> usize {
         self.length.try_into().unwrap()
     }
-    pub fn end(&self) -> usize {
+    pub fn end(self) -> usize {
         self.start
             .checked_add(self.length)
             .expect("Overflow")
             .try_into()
             .unwrap()
     }
-    pub fn combine(&self, other: Self) -> Self {
+    pub fn combine(self, other: Self) -> Self {
         let start = std::cmp::min(self.start, other.start);
         SourceRange {
             start,
@@ -171,7 +173,7 @@ impl SourceRange {
     }
 
     // Return true if a location is in this range, including one-past-the-end.
-    pub fn contains_inclusive(&self, loc: usize) -> bool {
+    pub fn contains_inclusive(self, loc: usize) -> bool {
         self.start() <= loc && loc - self.start() <= self.length()
     }
 }
@@ -198,6 +200,8 @@ impl ParseTokenType {
             ParseTokenType::background => L!("ParseTokenType::background"),
             ParseTokenType::end => L!("ParseTokenType::end"),
             ParseTokenType::pipe => L!("ParseTokenType::pipe"),
+            ParseTokenType::left_brace => L!("ParseTokenType::lbrace"),
+            ParseTokenType::right_brace => L!("ParseTokenType::rbrace"),
             ParseTokenType::redirection => L!("ParseTokenType::redirection"),
             ParseTokenType::string => L!("ParseTokenType::string"),
             ParseTokenType::andand => L!("ParseTokenType::andand"),
@@ -210,7 +214,7 @@ impl ParseTokenType {
 
 impl Default for ParseKeyword {
     fn default() -> Self {
-        ParseKeyword::none
+        ParseKeyword::None
     }
 }
 
@@ -218,24 +222,24 @@ impl ParseKeyword {
     /// Return the keyword as a string.
     pub fn to_wstr(self) -> &'static wstr {
         match self {
-            ParseKeyword::kw_exclam => L!("!"),
-            ParseKeyword::kw_and => L!("and"),
-            ParseKeyword::kw_begin => L!("begin"),
-            ParseKeyword::kw_builtin => L!("builtin"),
-            ParseKeyword::kw_case => L!("case"),
-            ParseKeyword::kw_command => L!("command"),
-            ParseKeyword::kw_else => L!("else"),
-            ParseKeyword::kw_end => L!("end"),
-            ParseKeyword::kw_exec => L!("exec"),
-            ParseKeyword::kw_for => L!("for"),
-            ParseKeyword::kw_function => L!("function"),
-            ParseKeyword::kw_if => L!("if"),
-            ParseKeyword::kw_in => L!("in"),
-            ParseKeyword::kw_not => L!("not"),
-            ParseKeyword::kw_or => L!("or"),
-            ParseKeyword::kw_switch => L!("switch"),
-            ParseKeyword::kw_time => L!("time"),
-            ParseKeyword::kw_while => L!("while"),
+            ParseKeyword::And => L!("and"),
+            ParseKeyword::Begin => L!("begin"),
+            ParseKeyword::Builtin => L!("builtin"),
+            ParseKeyword::Case => L!("case"),
+            ParseKeyword::Command => L!("command"),
+            ParseKeyword::Else => L!("else"),
+            ParseKeyword::End => L!("end"),
+            ParseKeyword::Exclam => L!("!"),
+            ParseKeyword::Exec => L!("exec"),
+            ParseKeyword::For => L!("for"),
+            ParseKeyword::Function => L!("function"),
+            ParseKeyword::If => L!("if"),
+            ParseKeyword::In => L!("in"),
+            ParseKeyword::Not => L!("not"),
+            ParseKeyword::Or => L!("or"),
+            ParseKeyword::Switch => L!("switch"),
+            ParseKeyword::Time => L!("time"),
+            ParseKeyword::While => L!("while"),
             _ => L!("unknown_keyword"),
         }
     }
@@ -249,26 +253,28 @@ impl fish_printf::ToArg<'static> for ParseKeyword {
 
 impl From<&wstr> for ParseKeyword {
     fn from(s: &wstr) -> Self {
-        match s {
-            _ if s == "!" => ParseKeyword::kw_exclam,
-            _ if s == "and" => ParseKeyword::kw_and,
-            _ if s == "begin" => ParseKeyword::kw_begin,
-            _ if s == "builtin" => ParseKeyword::kw_builtin,
-            _ if s == "case" => ParseKeyword::kw_case,
-            _ if s == "command" => ParseKeyword::kw_command,
-            _ if s == "else" => ParseKeyword::kw_else,
-            _ if s == "end" => ParseKeyword::kw_end,
-            _ if s == "exec" => ParseKeyword::kw_exec,
-            _ if s == "for" => ParseKeyword::kw_for,
-            _ if s == "function" => ParseKeyword::kw_function,
-            _ if s == "if" => ParseKeyword::kw_if,
-            _ if s == "in" => ParseKeyword::kw_in,
-            _ if s == "not" => ParseKeyword::kw_not,
-            _ if s == "or" => ParseKeyword::kw_or,
-            _ if s == "switch" => ParseKeyword::kw_switch,
-            _ if s == "time" => ParseKeyword::kw_time,
-            _ if s == "while" => ParseKeyword::kw_while,
-            _ => ParseKeyword::none,
+        // Note this is called in hot loops.
+        let c0 = s.as_char_slice().get(0).copied().unwrap_or('\0');
+        match c0 {
+            '!' if s == L!("!") => ParseKeyword::Exclam,
+            'a' if s == L!("and") => ParseKeyword::And,
+            'b' if s == L!("begin") => ParseKeyword::Begin,
+            'b' if s == L!("builtin") => ParseKeyword::Builtin,
+            'c' if s == L!("case") => ParseKeyword::Case,
+            'c' if s == L!("command") => ParseKeyword::Command,
+            'e' if s == L!("else") => ParseKeyword::Else,
+            'e' if s == L!("end") => ParseKeyword::End,
+            'e' if s == L!("exec") => ParseKeyword::Exec,
+            'f' if s == L!("for") => ParseKeyword::For,
+            'f' if s == L!("function") => ParseKeyword::Function,
+            'i' if s == L!("if") => ParseKeyword::If,
+            'i' if s == L!("in") => ParseKeyword::In,
+            'n' if s == L!("not") => ParseKeyword::Not,
+            'o' if s == L!("or") => ParseKeyword::Or,
+            's' if s == L!("switch") => ParseKeyword::Switch,
+            't' if s == L!("time") => ParseKeyword::Time,
+            'w' if s == L!("while") => ParseKeyword::While,
+            _ => ParseKeyword::None,
         }
     }
 }
@@ -409,7 +415,7 @@ pub fn token_type_user_presentable_description(
     type_: ParseTokenType,
     keyword: ParseKeyword,
 ) -> WString {
-    if keyword != ParseKeyword::none {
+    if keyword != ParseKeyword::None {
         return sprintf!("keyword: '%ls'", keyword.to_wstr());
     }
     match type_ {
@@ -417,6 +423,8 @@ pub fn token_type_user_presentable_description(
         ParseTokenType::pipe => L!("a pipe").to_owned(),
         ParseTokenType::redirection => L!("a redirection").to_owned(),
         ParseTokenType::background => L!("a '&'").to_owned(),
+        ParseTokenType::left_brace => L!("a '{'").to_owned(),
+        ParseTokenType::right_brace => L!("a '}'").to_owned(),
         ParseTokenType::andand => L!("'&&'").to_owned(),
         ParseTokenType::oror => L!("'||'").to_owned(),
         ParseTokenType::end => L!("end of the statement").to_owned(),
@@ -444,83 +452,95 @@ pub fn parse_error_offset_source_start(errors: &mut ParseErrorList, amt: usize)
 }
 
 /// Maximum number of function calls.
-pub const FISH_MAX_STACK_DEPTH: usize = 128;
+pub const FISH_MAX_STACK_DEPTH: isize = 128;
 
 /// Maximum number of nested string substitutions (in lieu of evals)
 /// Reduced under TSAN: our CI test creates 500 jobs and this is very slow with TSAN.
 #[cfg(feature = "tsan")]
-pub const FISH_MAX_EVAL_DEPTH: usize = 250;
+pub const FISH_MAX_EVAL_DEPTH: isize = 250;
 #[cfg(not(feature = "tsan"))]
-pub const FISH_MAX_EVAL_DEPTH: usize = 500;
+pub const FISH_MAX_EVAL_DEPTH: isize = 500;
 
-/// Error message on a function that calls itself immediately.
-pub const INFINITE_FUNC_RECURSION_ERR_MSG: &str =
-    "The function '%ls' calls itself immediately, which would result in an infinite loop.";
+localizable_consts!(
+    /// Error message on a function that calls itself immediately.
+    pub INFINITE_FUNC_RECURSION_ERR_MSG
+    "The function '%ls' calls itself immediately, which would result in an infinite loop."
 
-/// Error message on reaching maximum call stack depth.
-pub const CALL_STACK_LIMIT_EXCEEDED_ERR_MSG: &str =
-    "The call stack limit has been exceeded. Do you have an accidental infinite loop?";
+    /// Error message on reaching maximum call stack depth.
+    pub CALL_STACK_LIMIT_EXCEEDED_ERR_MSG
+    "The call stack limit has been exceeded. Do you have an accidental infinite loop?"
 
-/// Error message when encountering an unknown builtin name.
-pub const UNKNOWN_BUILTIN_ERR_MSG: &str = "Unknown builtin '%ls'";
+    /// Error message when encountering an unknown builtin name.
+    pub UNKNOWN_BUILTIN_ERR_MSG
+    "Unknown builtin '%ls'"
 
-/// Error message when encountering a failed expansion, e.g. for the variable name in for loops.
-pub const FAILED_EXPANSION_VARIABLE_NAME_ERR_MSG: &str = "Unable to expand variable name '%ls'";
+    /// Error message when encountering a failed expansion, e.g. for the variable name in for loops.
+    pub FAILED_EXPANSION_VARIABLE_NAME_ERR_MSG
+    "Unable to expand variable name '%ls'"
 
-/// Error message when encountering an illegal file descriptor.
-pub const ILLEGAL_FD_ERR_MSG: &str = "Illegal file descriptor in redirection '%ls'";
+    /// Error message when encountering an illegal file descriptor.
+    pub ILLEGAL_FD_ERR_MSG
+    "Illegal file descriptor in redirection '%ls'"
 
-/// Error message for wildcards with no matches.
-pub const WILDCARD_ERR_MSG: &str = "No matches for wildcard '%ls'. See `help wildcards-globbing`.";
+    /// Error message for wildcards with no matches.
+    pub WILDCARD_ERR_MSG
+    "No matches for wildcard '%ls'. See `help wildcards-globbing`."
 
-/// Error when using break outside of loop.
-pub const INVALID_BREAK_ERR_MSG: &str = "'break' while not inside of loop";
+    /// Error when using break outside of loop.
+    pub INVALID_BREAK_ERR_MSG
+    "'break' while not inside of loop"
 
-/// Error when using continue outside of loop.
-pub const INVALID_CONTINUE_ERR_MSG: &str = "'continue' while not inside of loop";
+    /// Error when using continue outside of loop.
+    pub INVALID_CONTINUE_ERR_MSG
+    "'continue' while not inside of loop"
 
-/// Error message when a command may not be in a pipeline.
-pub const INVALID_PIPELINE_CMD_ERR_MSG: &str = "The '%ls' command can not be used in a pipeline";
+    /// Error message when a command may not be in a pipeline.
+    pub INVALID_PIPELINE_CMD_ERR_MSG
+    "The '%ls' command can not be used in a pipeline"
 
-// Error messages. The number is a reminder of how many format specifiers are contained.
+    // Error messages. The number is a reminder of how many format specifiers are contained.
 
-/// Error for $^.
-pub const ERROR_BAD_VAR_CHAR1: &str = "$%lc is not a valid variable in fish.";
+    /// Error for $^.
+    pub ERROR_BAD_VAR_CHAR1
+    "$%lc is not a valid variable in fish."
 
-/// Error for ${a}.
-pub const ERROR_BRACKETED_VARIABLE1: &str =
-    "Variables cannot be bracketed. In fish, please use {$%ls}.";
+    /// Error for ${a}.
+    pub ERROR_BRACKETED_VARIABLE1
+    "Variables cannot be bracketed. In fish, please use {$%ls}."
 
-/// Error for "${a}".
-pub const ERROR_BRACKETED_VARIABLE_QUOTED1: &str =
-    "Variables cannot be bracketed. In fish, please use \"$%ls\".";
+    /// Error for "${a}".
+    pub ERROR_BRACKETED_VARIABLE_QUOTED1
+    "Variables cannot be bracketed. In fish, please use \"$%ls\"."
 
-/// Error issued on $?.
-pub const ERROR_NOT_STATUS: &str = "$? is not the exit status. In fish, please use $status.";
+    /// Error issued on $?.
+    pub ERROR_NOT_STATUS
+    "$? is not the exit status. In fish, please use $status."
 
-/// Error issued on $$.
-pub const ERROR_NOT_PID: &str = "$$ is not the pid. In fish, please use $fish_pid.";
+    /// Error issued on $$.
+    pub ERROR_NOT_PID
+    "$$ is not the pid. In fish, please use $fish_pid."
 
-/// Error issued on $#.
-pub const ERROR_NOT_ARGV_COUNT: &str = "$# is not supported. In fish, please use 'count $argv'.";
+    /// Error issued on $#.
+    pub ERROR_NOT_ARGV_COUNT
+    "$# is not supported. In fish, please use 'count $argv'."
 
-/// Error issued on $@.
-pub const ERROR_NOT_ARGV_AT: &str = "$@ is not supported. In fish, please use $argv.";
+    /// Error issued on $@.
+    pub ERROR_NOT_ARGV_AT
+    "$@ is not supported. In fish, please use $argv."
 
-/// Error issued on $*.
-pub const ERROR_NOT_ARGV_STAR: &str = "$* is not supported. In fish, please use $argv.";
+    /// Error issued on $*.
+    pub ERROR_NOT_ARGV_STAR
+    "$* is not supported. In fish, please use $argv."
 
-/// Error issued on $.
-pub const ERROR_NO_VAR_NAME: &str = "Expected a variable name after this $.";
+    /// Error issued on $.
+    pub ERROR_NO_VAR_NAME
+    "Expected a variable name after this $."
 
-/// Error message for Posix-style assignment: foo=bar.
-pub const ERROR_BAD_COMMAND_ASSIGN_ERR_MSG: &str =
-    "Unsupported use of '='. In fish, please use 'set %ls %ls'.";
+    /// Error message for Posix-style assignment: foo=bar.
+    pub ERROR_BAD_COMMAND_ASSIGN_ERR_MSG
+    "Unsupported use of '='. In fish, please use 'set %ls %ls'."
 
-/// Error message for a command like `time foo &`.
-pub const ERROR_TIME_BACKGROUND: &str =
-    "'time' is not supported for background jobs. Consider using 'command time'.";
-
-/// Error issued on { echo; echo }.
-pub const ERROR_NO_BRACE_GROUPING: &str =
-    "'{ ... }' is not supported for grouping commands. Please use 'begin; ...; end'";
+    /// Error message for a command like `time foo &`.
+    pub ERROR_TIME_BACKGROUND
+    "'time' is not supported for background jobs. Consider using 'command time'."
+);
