@@ -1,10 +1,10 @@
 // The fish parser. Contains functions for parsing and evaluating code.
 
-use crate::ast::{self, Ast, List, Node};
+use crate::ast::{self, Node};
 use crate::builtins::shared::STATUS_ILLEGAL_CMD;
 use crate::common::{
-    escape_string, scoped_push_replacer, CancelChecker, EscapeFlags, EscapeStringStyle,
-    FilenameRef, ScopeGuarding, PROFILING_ACTIVE,
+    escape_string, wcs2string, CancelChecker, EscapeFlags, EscapeStringStyle, FilenameRef,
+    ScopeGuarding, ScopedCell, ScopedRefCell, PROFILING_ACTIVE,
 };
 use crate::complete::CompletionList;
 use crate::env::{EnvMode, EnvStack, EnvStackSetResult, Environment, Statuses};
@@ -14,7 +14,7 @@ use crate::expand::{
 };
 use crate::fds::{open_dir, BEST_O_SEARCH};
 use crate::global_safety::RelaxedAtomicBool;
-use crate::input_common::terminal_protocols_disable_ifn;
+use crate::input_common::TerminalQuery;
 use crate::io::IoChain;
 use crate::job_group::MaybeJobId;
 use crate::operation_context::{OperationContext, EXPANSION_LIMIT_DEFAULT};
@@ -23,29 +23,30 @@ use crate::parse_constants::{
     SOURCE_LOCATION_UNKNOWN,
 };
 use crate::parse_execution::{EndExecutionReason, ExecutionContext};
+use crate::parse_tree::NodeRef;
 use crate::parse_tree::{parse_source, LineCounter, ParsedSourceRef};
 use crate::proc::{job_reap, JobGroupRef, JobList, JobRef, Pid, ProcStatus};
 use crate::signal::{signal_check_cancel, signal_clear_cancel, Signal};
-use crate::threads::assert_is_main_thread;
 use crate::util::get_time;
 use crate::wait_handle::WaitHandleStore;
-use crate::wchar::{wstr, WString, L};
-use crate::wutil::{perror, wgettext, wgettext_fmt};
+use crate::wchar::prelude::*;
+use crate::wchar_ext::WExt;
+use crate::wutil::perror;
 use crate::{function, FLOG};
 use libc::c_int;
 #[cfg(not(target_has_atomic = "64"))]
 use portable_atomic::AtomicU64;
 use std::cell::{Ref, RefCell, RefMut};
 use std::ffi::{CStr, OsStr};
+use std::fs::File;
+use std::io::Write;
 use std::num::NonZeroU32;
-use std::os::fd::{AsRawFd, OwnedFd, RawFd};
+use std::os::fd::OwnedFd;
 use std::rc::Rc;
 #[cfg(target_has_atomic = "64")]
 use std::sync::atomic::AtomicU64;
-use std::sync::{
-    atomic::{AtomicIsize, Ordering},
-    Arc,
-};
+use std::sync::Arc;
+use std::time::Duration;
 
 pub enum BlockData {
     Function {
@@ -217,21 +218,74 @@ impl ProfileItem {
     }
 }
 
+/// Data which is managed in a scoped fashion: is generally set for the duration of a block
+/// of code. Note this is stored in a Cell and so must be Copy.
+#[derive(Copy, Clone)]
+pub struct ScopedData {
+    /// The 'depth' of the fish call stack.
+    /// -1 means nothing is executing. 0 means we are running a top-level command.
+    /// Larger values indicate deeper nesting.
+    pub eval_level: isize,
+
+    /// Whether we are running a subshell command.
+    pub is_subshell: bool,
+
+    /// Whether we are running an event handler.
+    pub is_event: bool,
+
+    /// Whether we are currently interactive.
+    pub is_interactive: bool,
+
+    /// Whether the command line is closed for modification from fish script.
+    pub readonly_commandline: bool,
+
+    /// Whether to suppress fish_trace output. This occurs in the prompt, event handlers, and key
+    /// bindings.
+    pub suppress_fish_trace: bool,
+
+    /// The read limit to apply to captured subshell output, or 0 for none.
+    pub read_limit: usize,
+
+    /// Whether we are currently cleaning processes.
+    pub is_cleaning_procs: bool,
+
+    /// The internal job id of the job being populated, or 0 if none.
+    /// This supports the '--on-job-exit caller' feature.
+    pub caller_id: u64, // TODO should be InternalJobId
+}
+
+impl Default for ScopedData {
+    fn default() -> Self {
+        Self {
+            eval_level: -1,
+            is_subshell: false,
+            is_event: false,
+            readonly_commandline: false,
+            is_interactive: false,
+            suppress_fish_trace: false,
+            read_limit: 0,
+            is_cleaning_procs: false,
+            caller_id: 0,
+        }
+    }
+}
+
 /// Miscellaneous data used to avoid recursion and others.
 #[derive(Default)]
 pub struct LibraryData {
     /// The current filename we are evaluating, either from builtin source or on the command line.
     pub current_filename: Option<FilenameRef>,
 
-    /// A stack of fake values to be returned by builtin_commandline. This is used by the completion
+    /// A fake value to be returned by builtin_commandline. This is used by the completion
     /// machinery when wrapping: e.g. if `tig` wraps `git` then git completions need to see git on
     /// the command line.
-    pub transient_commandlines: Vec<WString>,
+    pub transient_commandline: Option<WString>,
 
     /// A file descriptor holding the current working directory, for use in openat().
     /// This is never null and never invalid.
     pub cwd_fd: Option<Arc<OwnedFd>>,
 
+    /// Variables supporting the "status" builtin.
     pub status_vars: StatusVars,
 
     /// A counter incremented every time a command executes.
@@ -259,30 +313,6 @@ pub struct LibraryData {
     /// Whether we called builtin_complete -C without parameter.
     pub builtin_complete_current_commandline: bool,
 
-    /// Whether we are currently cleaning processes.
-    pub is_cleaning_procs: bool,
-
-    /// The internal job id of the job being populated, or 0 if none.
-    /// This supports the '--on-job-exit caller' feature.
-    pub caller_id: u64, // TODO should be InternalJobId
-
-    /// Whether we are running a subshell command.
-    pub is_subshell: bool,
-
-    /// Whether we are running an event handler. This is not a bool because we keep count of the
-    /// event nesting level.
-    pub is_event: i32,
-
-    /// Whether we are currently interactive.
-    pub is_interactive: bool,
-
-    /// Whether the command line is closed for modification from fish script.
-    pub readonly_commandline: bool,
-
-    /// Whether to suppress fish_trace output. This occurs in the prompt, event handlers, and key
-    /// bindings.
-    pub suppress_fish_trace: bool,
-
     /// Whether we should break or continue the current loop.
     /// This is set by the 'break' and 'continue' commands.
     pub loop_status: LoopStatus,
@@ -296,9 +326,6 @@ pub struct LibraryData {
     /// Note this only exits up to the "current script boundary." That is, a call to exit within a
     /// 'source' or 'read' command will only exit up to that command.
     pub exit_current_script: bool,
-
-    /// The read limit to apply to captured subshell output, or 0 for none.
-    pub read_limit: usize,
 }
 
 impl LibraryData {
@@ -359,7 +386,10 @@ pub enum ParserStatusVar {
     count_,
 }
 
-pub type BlockId = usize;
+/// A newtype for the block index.
+/// This is the naive position in the block list.
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+pub struct BlockId(usize);
 
 /// Controls the behavior when fish itself receives a signal and there are
 /// no blocks on the stack.
@@ -374,9 +404,11 @@ pub enum CancelBehavior {
 }
 
 pub struct Parser {
+    pub interactive_initialized: RelaxedAtomicBool,
+
     /// A shared line counter. This is handed out to each execution context
     /// so they can communicate the line number back to this Parser.
-    line_counter: Rc<RefCell<LineCounter<ast::JobPipeline>>>,
+    line_counter: ScopedRefCell<LineCounter<ast::JobPipeline>>,
 
     /// The jobs associated with this parser.
     job_list: RefCell<JobList>,
@@ -390,14 +422,14 @@ pub struct Parser {
     /// indexes during recursive evaluation.
     block_list: RefCell<Vec<Block>>,
 
-    /// The 'depth' of the fish call stack.
-    pub eval_level: AtomicIsize,
-
     /// Set of variables for the parser.
-    pub variables: Rc<EnvStack>,
+    pub variables: EnvStack,
+
+    /// Data managed in a scoped fashion.
+    scoped_data: ScopedCell<ScopedData>,
 
     /// Miscellaneous library data.
-    library_data: RefCell<LibraryData>,
+    pub library_data: ScopedRefCell<LibraryData>,
 
     /// If set, we synchronize universal variables after external commands,
     /// including sending on-variable change events.
@@ -411,23 +443,31 @@ pub struct Parser {
 
     /// Global event blocks.
     pub global_event_blocks: AtomicU64,
+
+    pub blocking_query: RefCell<Option<TerminalQuery>>,
+
+    // Timeout for blocking terminal queries.
+    pub blocking_query_timeout: RefCell<Option<Duration>>,
 }
 
 impl Parser {
     /// Create a parser.
-    pub fn new(variables: Rc<EnvStack>, cancel_behavior: CancelBehavior) -> Parser {
+    pub fn new(variables: EnvStack, cancel_behavior: CancelBehavior) -> Parser {
         let result = Self {
-            line_counter: Rc::new(RefCell::new(LineCounter::empty())),
+            interactive_initialized: RelaxedAtomicBool::new(false),
+            line_counter: ScopedRefCell::new(LineCounter::empty()),
             job_list: RefCell::default(),
             wait_handles: RefCell::new(WaitHandleStore::new()),
             block_list: RefCell::default(),
-            eval_level: AtomicIsize::new(-1),
             variables,
-            library_data: RefCell::new(LibraryData::new()),
+            scoped_data: ScopedCell::new(ScopedData::default()),
+            library_data: ScopedRefCell::new(LibraryData::new()),
             syncs_uvars: RelaxedAtomicBool::new(false),
             cancel_behavior,
             profile_items: RefCell::default(),
             global_event_blocks: AtomicU64::new(0),
+            blocking_query: RefCell::new(None),
+            blocking_query_timeout: RefCell::new(None),
         };
 
         match open_dir(CStr::from_bytes_with_nul(b".\0").unwrap(), BEST_O_SEARCH) {
@@ -450,9 +490,7 @@ impl Parser {
 
     /// Return whether we are currently evaluating a function.
     pub fn is_function(&self) -> bool {
-        self.blocks()
-            .iter()
-            .rev()
+        self.blocks_iter_rev()
             // If a function sources a file, don't descend further.
             .take_while(|b| b.typ() != BlockType::source)
             .any(|b| b.is_function_call())
@@ -460,19 +498,12 @@ impl Parser {
 
     /// Return whether we are currently evaluating a command substitution.
     pub fn is_command_substitution(&self) -> bool {
-        self.blocks()
-            .iter()
-            .rev()
+        self.blocks_iter_rev()
             // If a function sources a file, don't descend further.
             .take_while(|b| b.typ() != BlockType::source)
             .any(|b| b.typ() == BlockType::subst)
     }
 
-    /// Assert that this parser is allowed to execute on the current thread.
-    pub fn assert_can_execute(&self) {
-        assert_is_main_thread();
-    }
-
     pub fn eval(&self, cmd: &wstr, io: &IoChain) -> EvalRes {
         self.eval_with(cmd, io, None, BlockType::top)
     }
@@ -509,10 +540,10 @@ impl Parser {
         eprintf!("%s\n", backtrace_and_desc);
 
         // Set a valid status.
-        self.set_last_statuses(Statuses::just(STATUS_ILLEGAL_CMD.unwrap()));
+        self.set_last_statuses(Statuses::just(STATUS_ILLEGAL_CMD));
         let break_expand = true;
         EvalRes {
-            status: ProcStatus::from_exit_code(STATUS_ILLEGAL_CMD.unwrap()),
+            status: ProcStatus::from_exit_code(STATUS_ILLEGAL_CMD),
             break_expand,
             ..Default::default()
         }
@@ -527,11 +558,11 @@ impl Parser {
         job_group: Option<&JobGroupRef>,
         block_type: BlockType,
     ) -> EvalRes {
-        assert!([BlockType::top, BlockType::subst].contains(&block_type));
-        let job_list = ps.ast.top().as_job_list().unwrap();
+        assert!(matches!(block_type, BlockType::top | BlockType::subst));
+        let job_list = ps.top_job_list();
         if !job_list.is_empty() {
             // Execute the top job list.
-            self.eval_node(ps, job_list, io, job_group, block_type)
+            self.eval_node(&job_list, io, job_group, block_type)
         } else {
             let status = ProcStatus::from_exit_code(self.get_last_status());
             EvalRes {
@@ -543,19 +574,64 @@ impl Parser {
         }
     }
 
+    pub fn eval_wstr(
+        &self,
+        src: WString,
+        io: &IoChain,
+        job_group: Option<&JobGroupRef>,
+        block_type: BlockType,
+    ) -> Result<EvalRes, WString> {
+        use crate::parse_tree::ParsedSource;
+        use crate::parse_util::parse_util_detect_errors_in_ast;
+        let mut errors = vec![];
+        let ast = ast::parse(&src, ParseTreeFlags::empty(), Some(&mut errors));
+        let mut errored = ast.errored();
+        if !errored {
+            errored = parse_util_detect_errors_in_ast(&ast, &src, Some(&mut errors)).is_err();
+        }
+        if errored {
+            let sb = self.get_backtrace(&src, &errors);
+            return Err(sb);
+        }
+
+        // Construct a parsed source ref.
+        // Be careful to transfer ownership, this could be a very large string.
+        let ps = Arc::new(ParsedSource::new(src, ast));
+        Ok(self.eval_parsed_source(&ps, io, job_group, block_type))
+    }
+
+    pub fn eval_file_wstr(
+        &self,
+        src: WString,
+        filename: Arc<WString>,
+        io: &IoChain,
+        job_group: Option<&JobGroupRef>,
+    ) -> Result<EvalRes, WString> {
+        let _interactive_push = self.push_scope(|s| s.is_interactive = false);
+        let sb = self.push_block(Block::source_block(filename.clone()));
+        let _filename_push = self
+            .library_data
+            .scoped_set(Some(filename), |s| &mut s.current_filename);
+
+        let ret = self.eval_wstr(src, io, job_group, BlockType::top);
+
+        self.pop_block(sb);
+        self.libdata_mut().exit_current_script = false;
+        ret
+    }
+
     /// Evaluates a node.
     /// The node type must be ast::Statement or ast::JobList.
     pub fn eval_node<T: Node>(
         &self,
-        ps: &ParsedSourceRef,
-        node: &T,
+        node: &NodeRef<T>,
         block_io: &IoChain,
         job_group: Option<&JobGroupRef>,
         block_type: BlockType,
     ) -> EvalRes {
         // Only certain blocks are allowed.
         assert!(
-            [BlockType::top, BlockType::subst].contains(&block_type),
+            matches!(block_type, BlockType::top | BlockType::subst),
             "Invalid block type"
         );
 
@@ -606,24 +682,20 @@ impl Parser {
         op_ctx.cancel_checker = cancel_checker;
 
         // Restore the line counter.
-        let line_counter = Rc::clone(&self.line_counter);
-        let scoped_line_counter =
-            scoped_push_replacer(|v| line_counter.replace(v), ps.line_counter());
+        let ps = node.parsed_source_ref();
+        let restore_line_counter = self.line_counter.scoped_replace(ps.line_counter());
 
         // Create a new execution context.
-        let mut execution_context =
-            ExecutionContext::new(ps.clone(), block_io.clone(), Rc::clone(&line_counter));
-
-        terminal_protocols_disable_ifn(false);
+        let mut execution_context = ExecutionContext::new(ps, block_io.clone(), &self.line_counter);
 
         // Check the exec count so we know if anything got executed.
         let prev_exec_count = self.libdata().exec_count;
         let prev_status_count = self.libdata().status_count;
-        let reason = execution_context.eval_node(&op_ctx, node, Some(scope_block));
+        let reason = execution_context.eval_node(&op_ctx, &**node, Some(scope_block));
         let new_exec_count = self.libdata().exec_count;
         let new_status_count = self.libdata().status_count;
 
-        ScopeGuarding::commit(scoped_line_counter);
+        ScopeGuarding::commit(restore_line_counter);
         self.pop_block(scope_block);
 
         job_reap(self, false); // reap again
@@ -652,7 +724,7 @@ impl Parser {
         ctx: &OperationContext<'_>,
     ) -> CompletionList {
         // Parse the string as an argument list.
-        let ast = Ast::parse_argument_list(arg_list_src, ParseTreeFlags::default(), None);
+        let ast = ast::parse_argument_list(arg_list_src, ParseTreeFlags::default(), None);
         if ast.errored() {
             // Failed to parse. Here we expect to have reported any errors in test_args.
             return vec![];
@@ -660,8 +732,7 @@ impl Parser {
 
         // Get the root argument list and extract arguments from it.
         let mut result = vec![];
-        let list = ast.top().as_freestanding_argument_list().unwrap();
-        for arg in &list.arguments {
+        for arg in &ast.top().arguments {
             let arg_src = arg.source(arg_list_src);
             if matches!(
                 expand_string(arg_src.to_owned(), &mut result, flags, ctx, None).result,
@@ -740,7 +811,7 @@ impl Parser {
     /// This supports 'status is-block'.
     pub fn is_block(&self) -> bool {
         // Note historically this has descended into 'source', unlike 'is_function'.
-        self.blocks().iter().rev().any(|b| {
+        self.blocks_iter_rev().any(|b| {
             ![
                 BlockType::top,
                 BlockType::subst,
@@ -752,37 +823,51 @@ impl Parser {
 
     /// Return whether we have a breakpoint block.
     pub fn is_breakpoint(&self) -> bool {
-        self.blocks()
-            .iter()
-            .rev()
+        self.blocks_iter_rev()
             .any(|b| b.typ() == BlockType::breakpoint)
     }
 
-    /// Return the list of blocks. The first block is at the top.
-    /// todo!("this RAII object should only be used for iterating over it (in reverse). Maybe enforce this")
-    pub fn blocks(&self) -> Ref<'_, Vec<Block>> {
-        self.block_list.borrow()
+    // Return an iterator over the blocks, in reverse order.
+    // That is, the first block is the innermost block.
+    pub fn blocks_iter_rev<'a>(&'a self) -> impl Iterator<Item = Ref<'a, Block>> {
+        let blocks = self.block_list.borrow();
+        let mut indices = (0..blocks.len()).rev();
+        std::iter::from_fn(move || {
+            let last = indices.next()?;
+            // note this clone is cheap
+            Some(Ref::map(Ref::clone(&blocks), |bl| &bl[last]))
+        })
     }
+
+    // Return the block at a given index, where 0 is the innermost block.
     pub fn block_at_index(&self, index: usize) -> Option<Ref<'_, Block>> {
-        let block_list = self.blocks();
-        if index >= block_list.len() {
+        let block_list = self.block_list.borrow();
+        let block_count = block_list.len();
+        if index >= block_count {
             None
         } else {
-            Some(Ref::map(block_list, |bl| &bl[bl.len() - 1 - index]))
+            Some(Ref::map(block_list, |bl| &bl[block_count - 1 - index]))
         }
     }
+
+    // Return the block at a given index, where 0 is the innermost block.
     pub fn block_at_index_mut(&self, index: usize) -> Option<RefMut<'_, Block>> {
         let block_list = self.block_list.borrow_mut();
-        if index >= block_list.len() {
+        let block_count = block_list.len();
+        if index >= block_count {
             None
         } else {
             Some(RefMut::map(block_list, |bl| {
-                let len = bl.len();
-                &mut bl[len - 1 - index]
+                &mut bl[block_count - 1 - index]
             }))
         }
     }
 
+    // Return the block with the given id, asserting it exists. Note ids are recycled.
+    pub fn block_with_id(&self, id: BlockId) -> Ref<'_, Block> {
+        Ref::map(self.block_list.borrow(), |bl| &bl[id.0])
+    }
+
     pub fn blocks_size(&self) -> usize {
         self.block_list.borrow().len()
     }
@@ -800,15 +885,28 @@ impl Parser {
         &self.variables
     }
 
-    /// Get the variables as an Arc.
-    pub fn vars_ref(&self) -> Rc<EnvStack> {
-        Rc::clone(&self.variables)
+    /// Get a copy of the scoped data.
+    #[inline(always)]
+    pub fn scope(&self) -> ScopedData {
+        self.scoped_data.get()
+    }
+
+    /// Modify the scoped values for the duration of the caller's scope (or whenever the ParserScope is dropped).
+    /// This accepts a closure which modifies the ScopedData, and returns a ParserScope which restores the
+    /// data when dropped.
+    pub fn push_scope<'a, F: FnOnce(&mut ScopedData)>(
+        &'a self,
+        modifier: F,
+    ) -> impl ScopeGuarding + 'a {
+        self.scoped_data.scoped_mod(modifier)
     }
 
     /// Get the library data.
     pub fn libdata(&self) -> Ref<'_, LibraryData> {
         self.library_data.borrow()
     }
+
+    /// Get the library data, mutably.
     pub fn libdata_mut(&self) -> RefMut<'_, LibraryData> {
         self.library_data.borrow_mut()
     }
@@ -874,14 +972,14 @@ impl Parser {
 
         let mut block_list = self.block_list.borrow_mut();
         block_list.push(block);
-        block_list.len() - 1
+        BlockId(block_list.len() - 1)
     }
 
     /// Remove the outermost block, asserting it's the given one.
     pub fn pop_block(&self, expected: BlockId) {
         let block = {
             let mut block_list = self.block_list.borrow_mut();
-            assert!(expected == block_list.len() - 1);
+            assert!(expected.0 == block_list.len() - 1);
             block_list.pop().unwrap()
         };
         if block.wants_pop_env() {
@@ -896,9 +994,7 @@ impl Parser {
             // isn't one return the function name for the current level.
             // Walk until we find a breakpoint, then take the next function.
             return self
-                .blocks()
-                .iter()
-                .rev()
+                .blocks_iter_rev()
                 .skip_while(|b| b.typ() != BlockType::breakpoint)
                 .find_map(|b| match b.data() {
                     Some(BlockData::Function { name, .. }) => Some(name.clone()),
@@ -906,9 +1002,7 @@ impl Parser {
                 });
         }
 
-        self.blocks()
-            .iter()
-            .rev()
+        self.blocks_iter_rev()
             // Historical: If we want the topmost function, but we are really in a file sourced by a
             // function, don't consider ourselves to be in a function.
             .take_while(|b| !(level == 1 && b.typ() == BlockType::source))
@@ -953,10 +1047,11 @@ impl Parser {
     }
 
     /// Returns the job and job index with the given pid.
+    /// This assumes that all external jobs have a pid.
     pub fn job_get_with_index_from_pid(&self, pid: Pid) -> Option<(usize, JobRef)> {
         for (i, job) in self.jobs().iter().enumerate() {
             for p in job.external_procs() {
-                if p.pid.load().unwrap() == pid {
+                if p.pid().unwrap() == pid {
                     return Some((i, job.clone()));
                 }
             }
@@ -966,7 +1061,7 @@ impl Parser {
 
     /// Returns a new profile item if profiling is active. The caller should fill it in.
     /// The Parser will deallocate it.
-    /// If profiling is not active, this returns nullptr.
+    /// If profiling is not active, this returns None.
     pub fn create_profile_item(&self) -> Option<usize> {
         if PROFILING_ACTIVE.load() {
             let mut profile_items = self.profile_items.borrow_mut();
@@ -989,7 +1084,7 @@ impl Parser {
     pub fn emit_profiling(&self, path: &OsStr) {
         // Save profiling information. OK to not use CLO_EXEC here because this is called while fish is
         // exiting (and hence will not fork).
-        let f = match std::fs::File::create(path) {
+        let mut f = match std::fs::File::create(path) {
             Ok(f) => f,
             Err(err) => {
                 FLOG!(
@@ -1003,8 +1098,7 @@ impl Parser {
                 return;
             }
         };
-        fprintf!(f.as_raw_fd(), "Time\tSum\tCommand\n");
-        print_profile(&self.profile_items.borrow(), f.as_raw_fd());
+        print_profile(&self.profile_items.borrow(), &mut f);
     }
 
     pub fn get_backtrace(&self, src: &wstr, errors: &ParseErrorList) -> WString {
@@ -1055,9 +1149,7 @@ impl Parser {
     /// reader_current_filename, e.g. if we are evaluating a function defined in a different file
     /// than the one currently read.
     pub fn current_filename(&self) -> Option<FilenameRef> {
-        self.blocks()
-            .iter()
-            .rev()
+        self.blocks_iter_rev()
             .find_map(|b| match b.data() {
                 Some(BlockData::Function { name, .. }) => {
                     function::get_props(name).and_then(|props| props.definition_file.clone())
@@ -1071,21 +1163,19 @@ impl Parser {
     /// Return if we are interactive, which means we are executing a command that the user typed in
     /// (and not, say, a prompt).
     pub fn is_interactive(&self) -> bool {
-        self.libdata().is_interactive
+        self.scope().is_interactive
     }
 
     /// Return a string representing the current stack trace.
     pub fn stack_trace(&self) -> WString {
-        self.blocks()
-            .iter()
-            .rev()
+        self.blocks_iter_rev()
             // Stop at event handler. No reason to believe that any other code is relevant.
             // It might make sense in the future to continue printing the stack trace of the code
             // that invoked the event, if this is a programmatic event, but we can't currently
             // detect that.
             .take_while(|b| b.typ() != BlockType::event)
             .fold(WString::new(), |mut trace, b| {
-                append_block_description_to_stack_trace(self, b, &mut trace);
+                append_block_description_to_stack_trace(self, &b, &mut trace);
                 trace
             })
     }
@@ -1095,18 +1185,15 @@ impl Parser {
         // We are interested in whether the count of functions on the stack exceeds
         // FISH_MAX_STACK_DEPTH. We don't separately track the number of functions, but we can have a
         // fast path through the eval_level. If the eval_level is in bounds, so must be the stack depth.
-        if self.eval_level.load(Ordering::Relaxed) <= isize::try_from(FISH_MAX_STACK_DEPTH).unwrap()
-        {
+        if self.scope().eval_level <= FISH_MAX_STACK_DEPTH {
             return false;
         }
         // Count the functions.
         let depth = self
-            .blocks()
-            .iter()
-            .rev()
+            .blocks_iter_rev()
             .filter(|b| b.is_function_call())
             .count();
-        depth > FISH_MAX_STACK_DEPTH
+        depth > (FISH_MAX_STACK_DEPTH as usize)
     }
 
     /// Mark whether we should sync universal variables.
@@ -1125,7 +1212,7 @@ impl Parser {
 
     /// Checks if the max eval depth has been exceeded
     pub fn is_eval_depth_exceeded(&self) -> bool {
-        self.eval_level.load(Ordering::Relaxed) >= isize::try_from(FISH_MAX_EVAL_DEPTH).unwrap()
+        self.scope().eval_level >= FISH_MAX_EVAL_DEPTH
     }
 }
 
@@ -1135,7 +1222,15 @@ fn user_presentable_path(path: &wstr, vars: &dyn Environment) -> WString {
 }
 
 /// Print profiling information to the specified stream.
-fn print_profile(items: &[ProfileItem], out: RawFd) {
+fn print_profile(items: &[ProfileItem], out: &mut File) {
+    let col_width = 10;
+    let _ = out.write_all(
+        format!(
+            "{:^col_width$} {:^col_width$} Command\n",
+            "Time (μs)", "Sum (μs)",
+        )
+        .as_bytes(),
+    );
     for (idx, item) in items.iter().enumerate() {
         if item.skipped || item.cmd.is_empty() {
             continue;
@@ -1144,7 +1239,7 @@ fn print_profile(items: &[ProfileItem], out: RawFd) {
         let total_time = item.duration;
 
         // Compute the self time as the total time, minus the total time consumed by subsequent
-        // items exactly one eval level deeper.
+        // items exactly one eval level deeper.
         let mut self_time = item.duration;
         for nested_item in items[idx + 1..].iter() {
             if nested_item.skipped {
@@ -1162,12 +1257,21 @@ fn print_profile(items: &[ProfileItem], out: RawFd) {
             }
         }
 
-        fprintf!(out, "%lld\t%lld\t", self_time, total_time);
-        for _i in 0..item.level {
-            fprintf!(out, "-");
-        }
-
-        fprintf!(out, "> %ls\n", item.cmd);
+        let level = item.level.unsigned_abs().saturating_add(1);
+        let _ = out.write_all(
+            format!(
+                "{:>col_width$} {:>col_width$} {:->level$} ",
+                self_time, total_time, '>'
+            )
+            .as_bytes(),
+        );
+        let indentation_level = col_width + 1 + col_width + 1 + level + 1;
+        let indented_cmd = item.cmd.replace(
+            L!("\n"),
+            &(WString::from("\n") + &wstr::repeat(L!(" "), indentation_level)[..]),
+        );
+        let _ = out.write_all(&wcs2string(&indented_cmd));
+        let _ = out.write_all(b"\n");
     }
 }
 
