@@ -3,7 +3,6 @@ use crate::input::{EventQueuePeeker, InputMappingSet, KeyNameStyle, DEFAULT_BIND
 use crate::input_common::{CharEvent, InputData, InputEventQueuer, KeyEvent};
 use crate::key::Key;
 use crate::wchar::prelude::*;
-use std::rc::Rc;
 
 struct TestInputEventQueuer {
     input_data: InputData,
@@ -20,9 +19,9 @@ impl InputEventQueuer for TestInputEventQueuer {
 
 #[test]
 fn test_input() {
-    let vars = Rc::new(EnvStack::new());
+    let vars = EnvStack::new();
     let mut input = TestInputEventQueuer {
-        input_data: InputData::new(i32::MAX), // value doesn't matter since we don't read from it
+        input_data: InputData::new(i32::MAX, None), // value doesn't matter since we don't read from it
     };
     // Ensure sequences are order independent. Here we add two bindings where the first is a prefix
     // of the second, and then emit the second key list. The second binding should be invoked, not
@@ -59,7 +58,7 @@ fn test_input() {
     }
 
     let mut peeker = EventQueuePeeker::new(&mut input);
-    let mapping = peeker.find_mapping(&*vars, &input_mappings);
+    let mapping = peeker.find_mapping(&vars, &input_mappings);
     assert!(mapping.is_some());
     assert!(mapping.unwrap().commands == ["down-line"]);
     peeker.restart();
