@@ -4,6 +4,7 @@
 use crate::flog::{FloggableDebug, FLOG};
 use crate::reader::Reader;
 use std::marker::PhantomData;
+use std::mem::MaybeUninit;
 use std::num::NonZeroU64;
 use std::sync::atomic::{AtomicBool, AtomicUsize, Ordering};
 use std::sync::{Arc, Mutex, OnceLock};
@@ -175,8 +176,8 @@ pub fn spawn<F: FnOnce() + Send + 'static>(callback: F) -> bool {
     // (#7837). Conservatively don't try to mask SIGKILL or SIGSTOP either; that's ignored on Linux
     // but maybe has an effect elsewhere.
     let saved_set = unsafe {
-        let mut new_set: libc::sigset_t = std::mem::zeroed();
-        let new_set = &mut new_set as *mut _;
+        let mut new_set = MaybeUninit::uninit();
+        let new_set = new_set.as_mut_ptr();
         libc::sigfillset(new_set);
         libc::sigdelset(new_set, libc::SIGILL); // bad jump
         libc::sigdelset(new_set, libc::SIGFPE); // divide-by-zero
@@ -206,7 +207,7 @@ pub fn spawn<F: FnOnce() + Send + 'static>(callback: F) -> bool {
             true
         }
         Err(e) => {
-            eprintln!("rust thread spawn failure: {e}");
+            eprintf!("rust thread spawn failure: %s\n", e);
             false
         }
     };
@@ -496,7 +497,9 @@ pub fn iothread_port() -> i32 {
 }
 
 pub fn iothread_service_main_with_timeout(ctx: &mut Reader, timeout: Duration) {
-    if crate::fd_readable_set::is_fd_readable(iothread_port(), timeout.as_millis() as u64) {
+    use crate::fd_readable_set::Timeout;
+
+    if crate::fd_readable_set::is_fd_readable(iothread_port(), Timeout::Duration(timeout)) {
         iothread_service_main(ctx);
     }
 }
@@ -531,7 +534,7 @@ pub(crate) fn iothread_drain_all(ctx: &mut Reader) {
     }
 }
 
-/// `Debounce` is a simple class which executes one function on a background thread while enqueing
+/// `Debounce` is a simple class which executes one function on a background thread while enqueuing
 /// at most one more. Subsequent execution requests overwrite the enqueued one. It takes an optional
 /// timeout; if a handler does not finish within the timeout then a new thread is spawned to service
 /// the remaining request.
@@ -551,7 +554,7 @@ pub struct Debounce {
 /// The data shared between [`Debounce`] instances.
 struct DebounceData {
     /// The (one or none) next enqueued request, overwritten each time a new call to
-    /// [`perform()`](Self::perform) is made.
+    /// [`Debounce::perform()`] is made.
     next_req: Option<WorkItem>,
     /// The non-zero token of the current non-abandoned thread or `None` if no thread is running.
     active_token: Option<NonZeroU64>,
@@ -607,9 +610,9 @@ impl Debounce {
         self.perform_with_completion(handler, |_ctx, _result| ())
     }
 
-    /// Enqueue `handler` to be performed on a background thread with [`Completion`] `completion`
-    /// to be performed on the main thread. If a function is already enqueued, this overwrites it
-    /// and that function will not be executed.
+    /// Enqueue `handler` to be performed on a background thread with `completion` to be performed
+    /// on the main thread. If a function is already enqueued, this overwrites it and that function
+    /// will not be executed.
     ///
     /// If the function executes within the optional timeout then `completion` will be invoked on
     /// the main thread with the result of the evaluated `handler`.
@@ -675,13 +678,13 @@ impl Debounce {
 }
 
 #[test]
-/// Verify that spawing a thread normally via [`std::thread::spawn()`] causes the calling thread's
+/// Verify that spawning a thread normally via [`std::thread::spawn()`] causes the calling thread's
 /// sigmask to be inherited by the newly spawned thread.
 fn std_thread_inherits_sigmask() {
     // First change our own thread mask
     let (saved_set, t1_set) = unsafe {
-        let mut new_set: libc::sigset_t = std::mem::zeroed();
-        let new_set = &mut new_set as *mut _;
+        let mut new_set = MaybeUninit::uninit();
+        let new_set = new_set.as_mut_ptr();
         libc::sigemptyset(new_set);
         libc::sigaddset(new_set, libc::SIGILL); // mask bad jump
 
@@ -691,8 +694,8 @@ fn std_thread_inherits_sigmask() {
 
         // Now get the current set that includes the masked SIGILL
         let mut t1_set: libc::sigset_t = std::mem::zeroed();
-        let mut empty_set = std::mem::zeroed();
-        let empty_set = &mut empty_set as *mut _;
+        let mut empty_set = MaybeUninit::uninit();
+        let empty_set = empty_set.as_mut_ptr();
         libc::sigemptyset(empty_set);
         let result = libc::pthread_sigmask(libc::SIG_UNBLOCK, empty_set, &mut t1_set as *mut _);
         assert_eq!(result, 0, "Failed to get own altered thread mask!");
@@ -704,8 +707,8 @@ fn std_thread_inherits_sigmask() {
     let t2_set = std::thread::scope(|_| {
         unsafe {
             // Set a new thread sigmask and verify that the old one is what we expect it to be
-            let mut new_set: libc::sigset_t = std::mem::zeroed();
-            let new_set = &mut new_set as *mut _;
+            let mut new_set = MaybeUninit::uninit();
+            let new_set = new_set.as_mut_ptr();
             libc::sigemptyset(new_set);
             let mut saved_set2: libc::sigset_t = std::mem::zeroed();
             let result = libc::pthread_sigmask(libc::SIG_BLOCK, new_set, &mut saved_set2 as *mut _);
