@@ -2,6 +2,8 @@
 //! and reclaiming it after.
 
 use crate::common::{self, safe_write_loop};
+use crate::env::Environment;
+use crate::env_dispatch::MIDNIGHT_COMMANDER_SID;
 use crate::flog::{FLOG, FLOGF};
 use crate::global_safety::RelaxedAtomicBool;
 use crate::job_group::JobGroup;
@@ -25,11 +27,6 @@ use std::sync::atomic::{AtomicBool, AtomicPtr, Ordering};
 /// Whether kitty keyboard protocol support is present in the TTY.
 static KITTY_KEYBOARD_SUPPORTED: OnceCell<bool> = OnceCell::new();
 
-/// Get whether the TTY supports the kitty keyboard protocol.
-pub fn get_kitty_keyboard_capability() -> Option<&'static bool> {
-    KITTY_KEYBOARD_SUPPORTED.get()
-}
-
 /// Set that the TTY supports the kitty keyboard protocol.
 pub fn maybe_set_kitty_keyboard_capability() {
     KITTY_KEYBOARD_SUPPORTED.get_or_init(|| true);
@@ -56,26 +53,36 @@ pub fn xtversion() -> Option<&'static wstr> {
     XTVERSION.get().as_ref().map(|s| s.as_utfstr())
 }
 
-// Facts about our environment, which inform how we handle the tty.
-#[derive(Debug, Copy, Clone)]
-pub struct TtyMetadata {
+// Facts that affect how we communicate with the TTY.
+#[derive(Debug, Copy, Clone, Eq, PartialEq)]
+pub enum TtyQuirks {
+    None,
+    // Running Midnight Commander which can't parse CSI yet.
+    PreCsiMidnightCommander,
+    // Running in iTerm2 before 3.5.12, which causes issues when using the kitty keyboard protocol.
+    PreKittyIterm2,
     // Whether we are running under tmux.
-    pub in_tmux: bool,
-
-    // If set, we are running before iTerm2 3.5.12, which does not support CSI-u.
-    pub pre_kitty_iterm2: bool,
+    Tmux,
+    // Whether we are running under WezTerm.
+    Wezterm,
 }
 
-impl TtyMetadata {
-    // Create a new TtyMetadata instance with the current environment.
-    fn detect(xtversion: &wstr) -> Self {
-        let in_tmux = xtversion.starts_with(L!("tmux "));
-        // Detect iTerm2 before 3.5.12.
-        let pre_kitty_iterm2 = get_iterm2_version(xtversion).is_some_and(|v| v < (3, 5, 12));
-
-        Self {
-            in_tmux,
-            pre_kitty_iterm2,
+impl TtyQuirks {
+    // Create a new TtyQuirks instance with the current environment.
+    fn detect(vars: &dyn Environment, xtversion: &wstr) -> Self {
+        use TtyQuirks::*;
+        if vars.get(MIDNIGHT_COMMANDER_SID).is_some()
+            && vars.get(L!("__mc_kitty_keyboard")).is_none()
+        {
+            PreCsiMidnightCommander
+        } else if get_iterm2_version(xtversion).is_some_and(|v| v < (3, 5, 12)) {
+            PreKittyIterm2
+        } else if xtversion.starts_with(L!("tmux ")) {
+            Tmux
+        } else if xtversion.starts_with(L!("WezTerm ")) {
+            Wezterm
+        } else {
+            None
         }
     }
 }
@@ -85,7 +92,8 @@ impl TtyMetadata {
 enum ProtocolKind {
     KittyKeyboard, // Kitty keyboard support, producing CSI-u style encoding.
     Other,         // Other protocols (e.g., modifyOtherKeys)
-    None,          // No protocols
+    WorkAroundWezTerm,
+    None, // No protocols
 }
 
 // Commands to emit to enable or disable TTY protocols. Each of these contains
@@ -95,6 +103,7 @@ enum ProtocolKind {
 struct ProtocolBytes {
     kitty_keyboard: Box<[u8]>,
     other: Box<[u8]>,
+    wezterm_workaround: Box<[u8]>,
     none: Box<[u8]>,
 }
 
@@ -102,19 +111,19 @@ struct ProtocolBytes {
 // This is created once at startup and then leaked, so it may be used
 // from the SIGTERM handler.
 struct TtyProtocolsSet {
-    // TTY metadata.
-    md: TtyMetadata,
+    // TTY quirks.
+    quirks: TtyQuirks,
     // Variants to enable or disable tty protocols.
     enablers: ProtocolBytes,
     disablers: ProtocolBytes,
 }
 
 impl TtyProtocolsSet {
-    // Get commands to enable or disable TTY protocols, based on the metadata
+    // Get commands to enable or disable TTY protocols
     // and the KITTY_KEYBOARD_SUPPORTED global variable.
     // THIS IS USED FROM A SIGNAL HANDLER.
     fn safe_get_commands(&self, enable: bool) -> &[u8] {
-        let protocol = self.md.safe_get_supported_protocol();
+        let protocol = self.quirks.safe_get_supported_protocol();
         let cmds = if enable {
             &self.enablers
         } else {
@@ -123,6 +132,7 @@ impl TtyProtocolsSet {
         match protocol {
             ProtocolKind::KittyKeyboard => &cmds.kitty_keyboard,
             ProtocolKind::Other => &cmds.other,
+            ProtocolKind::WorkAroundWezTerm => &cmds.wezterm_workaround,
             ProtocolKind::None => &cmds.none,
         }
     }
@@ -137,17 +147,26 @@ fn serialize_commands<'a>(cmds: impl Iterator<Item = TerminalCommand<'a>>) -> Bo
     out.contents().into()
 }
 
-impl TtyMetadata {
-    // Determine which keyboard protocol to use based on the metadata
-    // and the KITTY_KEYBOARD_SUPPORTED global variable.
+impl TtyQuirks {
+    // Determine which keyboard protocol.
     // This is used from a signal handler.
     fn safe_get_supported_protocol(&self) -> ProtocolKind {
-        if self.pre_kitty_iterm2 {
+        use TtyQuirks::{PreCsiMidnightCommander, PreKittyIterm2, Wezterm};
+        if *self == PreCsiMidnightCommander {
+            return ProtocolKind::None;
+        }
+        if *self == PreKittyIterm2 {
             return ProtocolKind::Other;
         }
         match KITTY_KEYBOARD_SUPPORTED.get() {
             Some(&true) => ProtocolKind::KittyKeyboard,
-            Some(&false) => ProtocolKind::Other,
+            Some(&false) => {
+                if *self == Wezterm {
+                    ProtocolKind::WorkAroundWezTerm
+                } else {
+                    ProtocolKind::Other
+                }
+            }
             None => ProtocolKind::None,
         }
     }
@@ -155,8 +174,13 @@ impl TtyMetadata {
     // Return the protocols set to enable or disable TTY protocols.
     fn get_protocols(self) -> TtyProtocolsSet {
         // Enable focus reporting under tmux
-        let focus_reporting_on = || self.in_tmux.then_some(DecsetFocusReporting).into_iter();
-        let focus_reporting_off = || self.in_tmux.then_some(DecrstFocusReporting).into_iter();
+        let (focus_reporting_on, focus_reporting_off) = {
+            let is_tmux = self == TtyQuirks::Tmux;
+            (
+                move || is_tmux.then_some(DecsetFocusReporting).into_iter(),
+                move || is_tmux.then_some(DecrstFocusReporting).into_iter(),
+            )
+        };
         let maybe_enable_focus_reporting = |protocols: &'static [TerminalCommand<'static>]| {
             protocols.iter().cloned().chain(focus_reporting_on())
         };
@@ -173,6 +197,10 @@ impl TtyMetadata {
                 ModifyOtherKeysEnable,       // XTerm's modifyOtherKeys
                 ApplicationKeypadModeEnable, // set application keypad mode, so the keypad keys send unique codes
             ])),
+            wezterm_workaround: serialize_commands(maybe_enable_focus_reporting(&[
+                DecsetBracketedPaste,
+                ApplicationKeypadModeEnable, // set application keypad mode, so the keypad keys send unique codes
+            ])),
             none: serialize_commands(maybe_enable_focus_reporting(&[DecsetBracketedPaste])),
         };
         let disablers = ProtocolBytes {
@@ -185,10 +213,14 @@ impl TtyMetadata {
                 ModifyOtherKeysDisable,
                 ApplicationKeypadModeDisable,
             ])),
+            wezterm_workaround: serialize_commands(maybe_disable_focus_reporting(&[
+                DecrstBracketedPaste,
+                ApplicationKeypadModeDisable,
+            ])),
             none: serialize_commands(maybe_disable_focus_reporting(&[DecrstBracketedPaste])),
         };
         TtyProtocolsSet {
-            md: self,
+            quirks: self,
             enablers,
             disablers,
         }
@@ -205,9 +237,8 @@ fn tty_protocols() -> Option<&'static TtyProtocolsSet> {
     unsafe { TTY_PROTOCOLS.load(Ordering::Acquire).as_ref() }
 }
 
-// Initialize TTY metadata.
-// This also initializes the terminal enable and disable serialized commands.
-pub fn initialize_tty_metadata() {
+// Initialize serialized commands for enabling/disabling TTY protocols in signal handlers.
+pub fn initialize_tty_protocols(vars: &dyn Environment) {
     // Default missing query responses.
     KITTY_KEYBOARD_SUPPORTED.get_or_init(|| false);
     SCROLL_CONTENT_UP_SUPPORTED.get_or_init(|| false);
@@ -218,7 +249,7 @@ pub fn initialize_tty_metadata() {
     let mut p = TTY_PROTOCOLS.load(Acquire);
     if p.is_null() {
         // Try to swap in a new TTY protocols set.
-        p = Box::into_raw(Box::new(TtyMetadata::detect(xtversion).get_protocols()));
+        p = Box::into_raw(Box::new(TtyQuirks::detect(vars, xtversion).get_protocols()));
         if let Err(_e) = TTY_PROTOCOLS.compare_exchange(std::ptr::null_mut(), p, Release, Acquire) {
             // Safety: p comes from Box::into_raw right above,
             // and wasn't shared with any other thread.
@@ -266,9 +297,10 @@ fn set_tty_protocols_active(on_write: fn(), enable: bool) -> bool {
 
     // Flog any terminal protocol changes of interest.
     let mode = if enable { "Enabling" } else { "Disabling" };
-    match protocols.md.safe_get_supported_protocol() {
+    match protocols.quirks.safe_get_supported_protocol() {
         ProtocolKind::KittyKeyboard => FLOG!(reader, mode, "kitty keyboard protocol"),
         ProtocolKind::Other => FLOG!(reader, mode, "other extended keys"),
+        ProtocolKind::WorkAroundWezTerm => FLOG!(reader, mode, "wezterm; no modifyOtherKeys"),
         ProtocolKind::None => (),
     };
     (on_write)();
