@@ -20,8 +20,10 @@ use crate::wchar::{wstr, WString, L};
 use crate::wchar_ext::WExt;
 use crate::wcstringutil::{join_strings, wcs2string_callback};
 use errno::errno;
-pub use gettext::{wgettext, wgettext_fmt, wgettext_maybe_fmt, wgettext_str};
-use std::ffi::OsStr;
+pub use gettext::{
+    localizable_consts, localizable_string, wgettext, wgettext_fmt, LocalizableString,
+};
+use std::ffi::{CStr, OsStr};
 use std::fs::{self, canonicalize};
 use std::io::{self, Write};
 use std::os::unix::prelude::*;
@@ -29,7 +31,7 @@ use std::os::unix::prelude::*;
 pub use crate::wutil::printf::{eprintf, fprintf, printf, sprintf};
 
 pub use fileid::{
-    file_id_for_fd, file_id_for_path, file_id_for_path_narrow, DevInode, FileId, INVALID_FILE_ID,
+    file_id_for_file, file_id_for_path, file_id_for_path_narrow, DevInode, FileId, INVALID_FILE_ID,
 };
 pub use wcstoi::*;
 
@@ -89,16 +91,15 @@ pub fn perror(s: &str) {
         let _ = write!(stderr, "{s}: ");
     }
     let slice = unsafe {
-        let msg = libc::strerror(e) as *const u8;
-        let len = libc::strlen(msg as *const _);
-        std::slice::from_raw_parts(msg, len)
+        let msg = libc::strerror(e);
+        CStr::from_ptr(msg).to_bytes()
     };
     let _ = stderr.write_all(slice);
     let _ = stderr.write_all(b"\n");
 }
 
 pub fn perror_io(s: &str, e: &io::Error) {
-    eprintln!("{}: {}", s, e);
+    eprintf!("%s: %s\n", s, e);
 }
 
 /// Wide character version of getcwd().
@@ -350,7 +351,7 @@ mod path_cd_tests {
     fn relative_path() {
         let wd = L!("/home/user/");
         let path = L!("projects");
-        eprintln!("({}, {})", wd, path);
+        eprintf!("(%ls, %ls)\n", wd, path);
         assert_eq!(path_normalize_for_cd(wd, path), L!("/home/user/projects"));
     }
 
@@ -358,7 +359,7 @@ mod path_cd_tests {
     fn absolute_path() {
         let wd = L!("/home/user/");
         let path = L!("/etc");
-        eprintln!("({}, {})", wd, path);
+        eprintf!("(%ls, %ls)\n", wd, path);
         assert_eq!(path_normalize_for_cd(wd, path), L!("/etc"));
     }
 
@@ -366,7 +367,7 @@ mod path_cd_tests {
     fn parent_directory() {
         let wd = L!("/home/user/projects/");
         let path = L!("../docs");
-        eprintln!("({}, {})", wd, path);
+        eprintf!("(%ls, %ls)\n", wd, path);
         assert_eq!(path_normalize_for_cd(wd, path), L!("/home/user/docs"));
     }
 
@@ -374,7 +375,7 @@ mod path_cd_tests {
     fn current_directory() {
         let wd = L!("/home/user/");
         let path = L!("./");
-        eprintln!("({}, {})", wd, path);
+        eprintf!("(%ls, %ls)\n", wd, path);
         assert_eq!(path_normalize_for_cd(wd, path), L!("/home/user"));
     }
 
@@ -382,7 +383,7 @@ mod path_cd_tests {
     fn nested_parent_directory() {
         let wd = L!("/home/user/projects/");
         let path = L!("../../");
-        eprintln!("({}, {})", wd, path);
+        eprintf!("(%ls, %ls)\n", wd, path);
         assert_eq!(path_normalize_for_cd(wd, path), L!("/home"));
     }
 
@@ -390,7 +391,7 @@ mod path_cd_tests {
     fn complex_path() {
         let wd = L!("/home/user/projects/");
         let path = L!("./../other/projects/./.././../docs");
-        eprintln!("({}, {})", wd, path);
+        eprintf!("(%ls, %ls)\n", wd, path);
         assert_eq!(
             path_normalize_for_cd(wd, path),
             L!("/home/user/other/projects/./.././../docs")
@@ -401,7 +402,7 @@ mod path_cd_tests {
     fn root_directory() {
         let wd = L!("/");
         let path = L!("..");
-        eprintln!("({}, {})", wd, path);
+        eprintf!("(%ls, %ls)\n", wd, path);
         assert_eq!(path_normalize_for_cd(wd, path), L!("/.."));
     }
 
@@ -409,7 +410,7 @@ mod path_cd_tests {
     fn up_to_root_directory() {
         let wd = L!("/foo/");
         let path = L!("..");
-        eprintln!("({}, {})", wd, path);
+        eprintf!("(%ls, %ls)\n", wd, path);
         assert_eq!(path_normalize_for_cd(wd, path), L!("/"));
     }
 
@@ -417,7 +418,7 @@ mod path_cd_tests {
     fn empty_path() {
         let wd = L!("/home/user/");
         let path = L!("");
-        eprintln!("({}, {})", wd, path);
+        eprintf!("(%ls, %ls)\n", wd, path);
         assert_eq!(path_normalize_for_cd(wd, path), L!("/home/user/"));
     }
 
@@ -425,7 +426,7 @@ mod path_cd_tests {
     fn trailing_slash() {
         let wd = L!("/home/user/projects/");
         let path = L!("docs/");
-        eprintln!("({}, {})", wd, path);
+        eprintf!("(%ls, %ls)\n", wd, path);
         assert_eq!(
             path_normalize_for_cd(wd, path),
             L!("/home/user/projects/docs/")
@@ -595,16 +596,6 @@ pub fn fish_iswalnum(c: char) -> bool {
     !fish_reserved_codepoint(c) && !fish_is_pua(c) && c.is_alphanumeric()
 }
 
-extern "C" {
-    fn iswgraph(wc: libc::wchar_t) -> libc::c_int; // Technically it's wint_t
-}
-
-/// We need this because there are too many implementations that don't return the proper answer for
-/// some code points. See issue #3050.
-pub fn fish_iswgraph(c: char) -> bool {
-    !fish_reserved_codepoint(c) && (fish_is_pua(c) || unsafe { iswgraph(c as libc::wchar_t) } != 0)
-}
-
 pub fn fish_wcswidth(s: &wstr) -> isize {
     fallback::fish_wcswidth(s)
 }
